
/*
 *   Xfrisk - The classic board game for X
 *   Copyright (C) 1993-1999 Elan Feingold (elan@aetherworks.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   $Id: aiClientMain.c,v 1.32 2003/07/14 10:37:37 sketch Exp $
 */

static const char
    rcsid[] = "@(#)$Id: aiClientMain.c,v 1.32 2003/07/14 10:37:37 sketch Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <netinet/in.h>
#include <netdb.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <signal.h>

#include "types.h" 
#include "utils.h"
#include "network.h"
#include "client.h"
#include "riskgame.h"
#include "game.h"
#include "debug.h"
#include "aiClient.h"
#include "aiController.h"

/* Private functions */
void  AI_Start( Int32 argc, CString *argv );
void  AI_RecoverFailure( CString strReason, Int32 iNumFailures );
void  AI_MainLoop( void );

void  AI_Replicate( const Msgs *pMess,
                    Int32 iType,
                    Int32 iSource );

/* Dummies or stubs or fill-ins */
void    CBK_IncomingMessage( Msgs *pMess );
void    FatalError( CString strError, Int32 iExitValue );

/*
 * Local data
 */
CString strClientName;
Int32   iCommLink, iReply;
Int32   iSpeciesID;
Int32   iCurrentClient;
Int32   iCurrentPlayer;
Int32   iState;
Flag    fGameInitialized = FALSE;
Flag    fGameStarted = FALSE;
void   *pvContext[ MAX_PLAYERS ];

/*
 * External references
 */
extern CString  __strName;
extern CString  __strAuthor;
extern CString  __strVersion;
extern CString  __strDesc;
extern void  *( *__AI_Callback )( void *, Int32, void * );
extern Int32 iMoveMode;

/**
 * AI client main function
 */
int main( int argc, char **argv )
    {
    Int32   i;
    Msgs    msg;
    char    *progname = argv[ 0 ];
    int     ch;

    /* Check args */
    while ( ( ch = getopt( argc, argv, "hv" ) ) != -1 )
        {
        switch (ch)
            {
            case 'v':
                CLNT_Version( progname );
                break;
            case 'h':
            default:
                CLNT_Usage( progname );
            }
        }

    argc -= optind;
    argv += optind;

    if ( !argc )
        CLNT_Usage( progname );

    for ( i = 0; i != MAX_PLAYERS; i++ )
        {
        pvContext[ i ] = ( void * ) NULL;
        }

    /* Initialize everything */
    RISK_InitObject( AI_Replicate,
                     NULL,
                     NULL,
                     AI_RecoverFailure,
                     NULL );

    AI_Start( argc, argv );

    /* Let the server know that we're ready to go as soon as it is... */
    msg.msgId = MSG_STARTGAME;
    RISK_SendMessage( CLIENTS_GetCommOfClient( CLNT_GetThisClientID( ) ),
                      &msg );
    iState = STATE_FORTIFY;

    /* And we're off */
    AI_MainLoop( );

    return( 0 );
    }


/**
 * AI startup function
 */
void AI_Start( Int32 argc, CString *argv )
    {
    struct sockaddr_in  server;
    struct hostent     *hp;
    char                strHostName[ MAXHOSTNAMELEN + 1 ];
    Int32               i;
    Msgs               *pMessage;
    Msgs                msg;
    char                buf[ 256 ];

    UNUSED( argc );

    /*
     * We want to ignore this signal, because we don't want a I/O
     * failure to terminate the program.
     */
  
    signal( SIGPIPE, SIG_IGN );
  
    /* Create sockets */
    if ( ( iCommLink = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
        {
        FatalError( "Cannot create CommLink.", 1 );
        }

    /* Connect socket using name specified by command line */
    server.sin_family = AF_INET;
  
    if ( ( hp = gethostbyname( argv[ 0 ] ) ) == 0 )
        {
        snprintf( buf, sizeof( buf ),
                  "AI-CLIENT: The host `%s' is unknown.", argv[ 0 ] );
        FatalError( buf, 1 );
        }
  
    memcpy( &server.sin_addr, hp->h_addr, hp->h_length );
    server.sin_port = htons( RISK_PORT );
  
    /* The Protocol: 
     *
     * Connect to the server.  Then do the following:
     *
     *  Send MSG_HELLO.
     *                      Receive MSG_VERSION     --> OK!
     *                              Other           --> Protocol bogosity.
     *  Send MSG_REGISTERCLIENT if version is correct.
     *                      Receive MSG_CLIENTIDENT --> OK!
     *                              MSG_CLIENTEXIT  --> Server is full.
     *                              Other           --> Protocol bogosity.
     *  Send MSG_VERSION if version is incorrect.
     */

    if ( connect( iCommLink, ( struct sockaddr * ) &server, sizeof( server ) ) < 0 )
        {
        printf( "AI-CLIENT: Cannot connect to a Xfrisk server on `%s'.\n"
                "        One probably needs to be started by running\n"
                "        `friskserver' on the machine.\n", argv[ 0 ] );
        FatalError( NULL, 0 );
        }
  
    /* Send the greeting */
    printf( "AI-CLIENT: Establishing CommLink to server.\n" );
    msg.msgId = MSG_HELLO;
    RISK_SendMessage( iCommLink, &msg );

    /* Get the version and check it. */
    RISK_ReceiveMessage( iCommLink, &pMessage );

    if ( pMessage->msgId != MSG_VERSION )
        {
        printf( "AI-CLIENT: Server is not following protocol ( %d )!  Exiting.\n", 
                pMessage->msgId );
        FatalError( NULL, 0 );
        }

    if ( strcmp( pMessage->msgVersion.strVersion, VERSION ) )
        {
        printf( "AI-CLIENT: Version mismatch ( server running %s )!  Exiting.\n",
                pMessage->msgVersion.strVersion );
        FatalError( NULL, 0 );
        }
  
    NET_DeleteMessage( pMessage );

    /* Send the server name information through the sockets so it can ID
     * the sending clients, and discover the pair of sockets that belong
     * to each client.  The server sends back the client ID.
     */
  
    gethostname( strHostName, sizeof( strHostName ) );
    msg.msgId = MSG_REGISTERCLIENT;

    msg.msgRegisterClient.strClientAddress = 
        ( CString ) MEM_Alloc( strlen( strHostName ) + 32 );

    snprintf( msg.msgRegisterClient.strClientAddress,
              strlen( strHostName ) + 32,
              "%s[%lu]",
              strHostName,
              ( unsigned long ) getpid( ) );

    strClientName = msg.msgRegisterClient.strClientAddress;
    msg.msgRegisterClient.iClientType = CLIENT_AI;

    /* Send back the registration message */
    RISK_SendMessage( iCommLink, &msg );

    /* Get the species ID */
    printf( "AI-CLIENT: Waiting for server to send species ID..." );
    RISK_ReceiveMessage( iCommLink, &pMessage );
  
    if ( pMessage->msgId != MSG_SPECIESIDENT )
        {
        printf( "Server is not following protocol ( %d )!", pMessage->msgId );
        FatalError( "Protocol error.", -1 );
        }
    else
        {
        printf( "Done.\n" );  
        }

    iSpeciesID = pMessage->msgSpeciesIdent.iSpeciesID;
    NET_DeleteMessage( pMessage );

    /* Wait for the response */
    printf( "AI-CLIENT: Waiting for server to send client ID..." );
    RISK_ReceiveMessage( iCommLink, &pMessage );
  
    if ( pMessage->msgId == MSG_EXIT )
        {
        FatalError( "Can't connect, server is full -- I'm impressed!", -1 );
        }
    else if ( pMessage->msgId == MSG_CLIENTIDENT )
        {
        printf( "Done.\n" );  

        /* Set the ID of the client, and then set the sockets of the client */
        iCurrentClient = pMessage->msgClientIdent.iClientID;
        }
    else
        {
        printf( "Server is not following protocol ( %d )!", pMessage->msgId );
        FatalError( "Protocol error!", -1 );
        }

    /* Free up memory */
    NET_DeleteMessage( pMessage );

    /* Get the species ID */
    i = iSpeciesID;
    printf( "AI-CLIENT: Me species #%d.  You Jane.\n", i );

    /* Init the species; afterwards, indicate the species is finished. */
    RISK_SetNameOfSpecies( i, __strName );
    RISK_SetClientOfSpecies( i, CLNT_GetThisClientID( ) );
    RISK_SetAuthorOfSpecies( i, __strAuthor );
    RISK_SetVersionOfSpecies( i, __strVersion );
    RISK_SetDescriptionOfSpecies( i, __strDesc );
    RISK_SetAllocationStateOfSpecies( i, ALLOC_COMPLETE );

    /* Let the player do one-time initializations */
    __AI_Callback( NULL, AI_INIT_ONCE, ( void * ) iSpeciesID ); 
    }
     


/**
 * Returns client ID of 'this' client
 */
Int32 CLNT_GetThisClientID( void )
    {
	return iCurrentClient;
    }

/**
 * Dummy function for getting the communication link ID
 */
Int32 CLIENTS_GetCommOfClient( Int32 iThisClient )
    {
    UNUSED( iThisClient );

    return iCommLink;
    }


/**
 * Recovery function for distributed object update failures
 */
void AI_RecoverFailure( CString strReason, Int32 iFailureCommlink )
    {
    printf( "AI-CLIENT: Fatal Error ( %s ) on CommLink %d.  Not equipped to\n"
            "recover from failure yet.  I'm exiting.\n", 
            strReason, iFailureCommlink );

    exit( -1 );
    }


/**
 * Fatal error exit function
 */
void FatalError( CString strError, Int32 iExitValue )
    {
    if ( strError )
        {
        printf( "AI-CLIENT: Fatal error ( %s )\n", strError );
        }

    exit( iExitValue );
    }


/**
 * AI client main loop
 */
void AI_MainLoop( void )
    {
    Msgs   *pMessage;
    fd_set  fdSet;
    fd_set  fdInteresting;
    Int32   retval;

    FD_ZERO( &fdInteresting );
    FD_SET( CLIENTS_GetCommOfClient( CLNT_GetThisClientID( ) ),
            &fdInteresting );

    /* Initialize everything */
    AI_Init( );

    /* Loop forever */
    for ( ; ; )
        {
        memcpy( &fdSet,
                &fdInteresting,
                sizeof( fdSet ) );

        /* Wait for a message */
        retval = select( CLIENTS_GetCommOfClient( CLNT_GetThisClientID( ) ) + 1,
                         &fdSet,
                         NULL,
                         NULL,
                         NULL );

        /* Sanity check */
        if ( retval < 0 )
            {
            perror( "Select" );
            }
      
        /* Receive the message */
        retval = RISK_ReceiveMessage( CLIENTS_GetCommOfClient( CLNT_GetThisClientID( ) ),
                                      &pMessage );

        /* Sanity check */
        if ( retval == 0 )
            {
            AI_RecoverFailure( "Server failed.", -1 );
            }
     
        /* Process it */
        if ( pMessage != NULL )
            {
            CBK_IncomingMessage( pMessage );
            NET_DeleteMessage( pMessage );
            }
        }
    }


/**
 * Incoming message handler callback
 */
void CBK_IncomingMessage( Msgs *pMessage )
    {
    Msgs msg;
    MsgTurnNotify *pvMess = &( pMessage->msgTurnNotify );
    MsgMessagePacket *msgMessagePacket = &( pMessage->msgMessagePacket );
    Int32 iWinner;

    switch ( pMessage->msgId )
        {
        case MSG_TURNNOTIFY:
            /* Set the current player */
            iCurrentPlayer = pvMess->iPlayer;

            if ( !fGameInitialized )
                {
                fGameInitialized = TRUE;
                /* Let the player do initializations of game */
                __AI_Callback( NULL, AI_INIT_GAME, ( void * )iSpeciesID );
                }

            /* Check to see if it's this client. */
            if ( pvMess->iClient == CLNT_GetThisClientID( ) )
                {
                pvContext[ iCurrentPlayer ] = 
                    __AI_Callback( pvContext[ iCurrentPlayer ],
                                   AI_INIT_TURN,
                                   NULL );

                /* See if the fortification is done */
                if ( !fGameStarted && ( iState == STATE_FORTIFY ) && 
                     ( RISK_GetNumArmiesOfPlayer( iCurrentPlayer ) == 0 ) )
                    {
                    fGameStarted = TRUE;
                    }

                if ( fGameStarted )
                    {
                    /* Give the players the initial armies */
                    GAME_SetTurnArmiesOfPlayer( iCurrentPlayer );

                    if ( RISK_GetNumCardsOfPlayer( iCurrentPlayer ) >= 3 )
                        {
                        pvContext[ iCurrentPlayer ] = 
                            __AI_Callback( pvContext[ iCurrentPlayer ],
                                           AI_EXCHANGE_CARDS,
                                           NULL );
                        AI_CheckCards( );
                        }

                    if ( iState == STATE_REGISTER )
                        {
                        return;
                        }
                    iState = STATE_PLACE;

                    do
                        {
                        /* Send the commands */
                        pvContext[ iCurrentPlayer ] = 
                            __AI_Callback( pvContext[ iCurrentPlayer ],
                                           AI_PLACE,
                                           NULL );

                        if ( iState == STATE_ATTACK )
                            {
                            AI_CheckPlacement( );
                            pvContext[ iCurrentPlayer ] = 
                                __AI_Callback( pvContext[ iCurrentPlayer ], 
                                               AI_ATTACK,
                                               NULL );
                            }
                        }
                    while ( iState == STATE_PLACE );

                    if ( iState == STATE_ATTACK )
                        {
                        iState = STATE_MOVE;
                        pvContext[ iCurrentPlayer ] = 
                            __AI_Callback( pvContext[ iCurrentPlayer ], 
                                           AI_MOVE,
                                           NULL );
                        }
                    }
                else if ( iState == STATE_FORTIFY )
                    {
                    pvContext[ iCurrentPlayer ] = 
                        __AI_Callback( pvContext[ iCurrentPlayer ], 
                                       AI_FORTIFY, 
                                       ( void * ) 1 );

                    AI_CheckFortification( );
                    }

                if ( iState != STATE_REGISTER )
                    {
                    /* End the turn */
                    AI_EndTurn( );
                    }
                }
            break;

        case MSG_MESSAGEPACKET:
            if ( msgMessagePacket->iFrom >= FROM_SERVER )
                {
                Int32 iOldPlayer = iCurrentPlayer;

                if ( msgMessagePacket->iTo >= 0 )
                    {
                    iCurrentPlayer = msgMessagePacket->iTo;
                    }
                else
                    {
                    iCurrentPlayer = RISK_GetNthPlayerAtClient( CLNT_GetThisClientID( ), 0 );
                    }

                __AI_Callback( NULL, AI_SERVER_MESSAGE, 
                               ( void * )( msgMessagePacket->strMessage ) );
                iCurrentPlayer = iOldPlayer;
                }
            else if ( ( msgMessagePacket->iFrom >= 0 ) &&
                      ( msgMessagePacket->iTo   >= 0 ) )
                {
                Int32 iOldPlayer = iCurrentPlayer;

                iCurrentPlayer = msgMessagePacket->iTo;
                pvContext[ iCurrentPlayer ] = 
                    __AI_Callback( pvContext[ iCurrentPlayer ],
                                   AI_MESSAGE, 
                                   ( void * )msgMessagePacket );
                iCurrentPlayer = iOldPlayer;
                }
            break;

        case MSG_FORCEEXCHANGECARDS:
            if ( pMessage->msgForceExchangeCards.iPlayer >= 0 )
                {
                if ( iState != STATE_ATTACK )
                    {
                    return;
                    }

                iCurrentPlayer = pMessage->msgForceExchangeCards.iPlayer;
                pvContext[ iCurrentPlayer ] = 
                    __AI_Callback( pvContext[ iCurrentPlayer ],
                                   AI_EXCHANGE_CARDS,
                                   NULL );
                AI_CheckCards( );

                if ( iState != STATE_ATTACK )
                    {
                    return;
                    }
                iState = STATE_PLACE;
                }
            break;

        case MSG_ENDOFMISSION:
            assert( iWinner >= -1 && iWinner < MAX_PLAYERS );

            iWinner = pMessage->msgEndOfMission.iWinner;
            if ( RISK_GetClientOfPlayer( iWinner ) == CLNT_GetThisClientID( ) )
                {
                if ( RISK_GetNumLivePlayers( ) <= 1 )
                    {
                    fGameStarted = FALSE;
                    iState = STATE_REGISTER;
                    msg.msgId = MSG_ENDOFGAME;
                    RISK_SendMessage( CLIENTS_GetCommOfClient( CLNT_GetThisClientID( ) ), 
                                      &msg );
                    }
                }
            break;

        case MSG_VICTORY:
            assert( iWinner >= -1 && iWinner < MAX_PLAYERS );

            iWinner = pMessage->msgVictory.iWinner;
            if ( RISK_GetClientOfPlayer( iWinner ) == CLNT_GetThisClientID( ) )
                {
                fGameStarted = FALSE;
                iState = STATE_REGISTER;
                msg.msgId = MSG_ENDOFGAME;
                RISK_SendMessage( CLIENTS_GetCommOfClient( CLNT_GetThisClientID( ) ), 
                                  &msg );
                }
            break;

        case MSG_ENDOFGAME:
            fGameStarted = FALSE;
            iState = STATE_FORTIFY;

            /* Let the server know that this client is ready to play */
            msg.msgId = MSG_STARTGAME;
            RISK_SendMessage( CLIENTS_GetCommOfClient( CLNT_GetThisClientID( ) ), 
                              &msg );

            fGameInitialized = FALSE;
            AI_Init( );
            break;

        case MSG_NOMESSAGE:
            break;

        case MSG_REPLYPACKET:
            iReply = pMessage->msgReplyPacket.iReply;
            break;

        case MSG_EXIT:
            UTIL_ExitProgram( 0 );
            break;

        default:
            break;
        }
    }


/**
 * Determine the number of armies to move
 */
Int32 UTIL_GetArmyNumber( Int32 iMin, Int32 iMax, Flag fLetCancel ) 
    {
    UNUSED( fLetCancel );

    if ( iMoveMode == ARMIES_MOVE_MAX )
        {
        return iMax;
        }

    if ( iMoveMode == ARMIES_MOVE_MIN )
        {
        return iMin;
        }

    if ( iMoveMode == ARMIES_MOVE_MANUAL )
        {
        Int32 iNumArmies = iMax - iMin;

        pvContext[ iCurrentPlayer ] =
            __AI_Callback( pvContext[ iCurrentPlayer ], AI_MOVE_MANUAL, 
                           ( void * ) &iNumArmies );
        return iNumArmies + iMin;
        }

    return ( iMax - iMin ) / 2;
    }


/**
 * Object state replication function
 */
void AI_Replicate( const Msgs *pMess, Int32 iType, Int32 iSource )
    {
    UNUSED( iSource );

    /*
     * Do the physical replication ( the server will broadcast it ),
     * but only if this callback is being called for an outgoing
     * message ( i.e. a replication ).  Otherwise, a message just
     * came in.
     */
    if ( iType == MESS_OUTGOING )
        {
        RISK_SendMessage( CLIENTS_GetCommOfClient( CLNT_GetThisClientID( ) ),
                          pMess );
        }
    }

/* End of file */
