/*
 *   Xfrisk - The classic board game for X
 *   Copyright (C) 1993-1999 Elan Feingold (elan@aetherworks.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   $Id: client.c,v 1.29 2003/07/14 10:37:38 sketch Exp $
 */

/** \file
 * Client functions and data, mainly about communication
 */

static const char
    rcsid[] = "@(#)$Id: client.c,v 1.29 2003/07/14 10:37:38 sketch Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <time.h>

#include "utils.h"
#include "network.h"
#include "client.h"
#include "callbacks.h"
#include "riskgame.h"
#include "gui-func.h"
#include "types.h"
#include "dice.h"
#include "debug.h"
#include "registerPlayers.h"
#include "viewStats.h"
#include "addPlayer.h"
#include "language.h"

/**
 * Client-specific information
 */
typedef struct
    {
    Int32   iCommLink;                              /**< just this commlink now */
    } Client;

static  Client  pClients[ MAX_CLIENTS ];            /**< to keep track of others? */
static  Flag    piCountryLightCount[ NUM_COUNTRIES ];
static  CString strClientName;
static  Int32   iCurrentClient;

/* Do something with these... */
void    FatalError( CString strError, Int32 iRetVal );


/**
 * Sets up communication stuff
 */
void CLNT_Init( int argc, char **argv )
    {
    struct sockaddr_in  server;
    struct hostent     *hp;
    struct timeval      tv;
    char                strHostName[MAXHOSTNAMELEN + 1];
    Int32               iCommLink;
    Int32               i;
    Msgs               *pMessage;
    Msgs                msg;
    char                buf[ 256 ];
    int                 len;

    UNUSED( argc );

    /* Seed the random number generator */
    gettimeofday( &tv, NULL );
    srand( tv.tv_sec * tv.tv_usec +
           ( getppid( ) * getpid( ) * getsid( getpid( ) ) ) );

    /* Init the client structures */
    for ( i = 0; i != NUM_COUNTRIES; i++ )
        {
        CLNT_SetLightCountOfCountry( i, 0 );
        }

    for ( i = 0; i != MAX_CLIENTS; i++ )
        {
        CLIENTS_SetCommLink( i, -1 );
        }

    /* We want to ignore this signal, because we don't want a I/O
     * failure to terminate the program.
     */
    signal( SIGPIPE, SIG_IGN );

    /* Create Commlink */
    if ( ( iCommLink = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
        {
        FatalError( ERR_COMMLINK, 1 );
        }

    /* Connect socket using name specified by command line */
    server.sin_family = AF_INET;

    if ( ( hp = gethostbyname( argv[ 0 ] ) ) == 0 )
        {
        snprintf( buf,
                  sizeof( buf ),
                  ERR_UNKNOWN_HOST,
                  argv[ 0 ] );

        FatalError( buf, 1 );
        }

    memcpy( &server.sin_addr, hp->h_addr, hp->h_length );
    server.sin_port = htons( RISK_PORT );

    /* Send the server name information through the sockets so it can ID
     * the sending clients, and discover the pair of sockets that belong
     * to each client.  The server sends back the client ID.
     */

    gethostname( strHostName, sizeof( strHostName ) );
    len = strlen( strHostName ) + 32;
    msg.msgRegisterClient.strClientAddress = ( CString ) MEM_Alloc( len );

    snprintf( msg.msgRegisterClient.strClientAddress,
              len,
              "%s[ %lu ]",
              strHostName,
              ( unsigned long ) getpid( ) );

    strClientName = msg.msgRegisterClient.strClientAddress;

    /* Connect to the server.  Connect the socket, send
     * MSG_REGISTER and then wait for a MSG_CLIENTIDENT
     * message to be send back.
     */

    if ( connect( iCommLink,
                  ( struct sockaddr * ) &server,
                  sizeof( server ) ) < 0 )
        {
        printf( ERR_COULD_NOT_CONNECT, argv[ 0 ] );
        FatalError( NULL, 0 );
        }

    /* Set the options on the CommLink */
    NET_SetCommLinkOptions( iCommLink );

    msg.msgId = MSG_OLDREGISTERCLIENT;
    RISK_SendMessage( iCommLink, &msg );
    printf( MSG_CONNECTED );
    printf( MSG_WAITING_SRV_ID );
    fflush( stdout );
    RISK_ReceiveMessage( iCommLink, &pMessage );
    printf( MSG_DONE );

    if ( pMessage->msgId == MSG_EXIT )
        {
        FatalError( ERR_SERVER_FULL, -1 );
        }
    else if ( pMessage->msgId == MSG_CLIENTIDENT )
        {
        /* Set the ID of the client, and then set the sockets of the client */
        iCurrentClient = pMessage->msgClientIdent.iClientID;
        CLIENTS_SetCommLink( CLNT_GetThisClientID( ), iCommLink );
        }
    else
        {
        FatalError( ERR_PROTOCOL_MISMATCH, -1 );
        }

    /* Free up memory */
    NET_DeleteMessage( pMessage );
    }


/**
 *
 */
void CLNT_PreViewVector( const Msgs *pMess )
    {
    STAT_Callback( pMess );
    }


/**
 *
 */
void CLNT_PostViewVector( const Msgs*pMess )
    {
    PLAYER_Callback( pMess );
    REG_Callback( pMess );
    CBK_Callback( pMess );
    }


/**
 *
 */
void CLNT_RecoverFailure( CString strReason, Int32 iCommLink )
    {
    UNUSED( iCommLink );

    printf( ERR_SERVER_FAILED, strReason );
    printf( ERR_NONRECOVERABLE );

    ( void ) UTIL_PopupDialog( ERR_OBJECT_FAILURE,
                               ERR_SRV_NONRECOVERABLE,
                               1,
                               "Ok",
                               NULL,
                               NULL );

    exit( 0 );
    }


/**
 * Allocate a player
 */
Int32 CLNT_AllocPlayer( void ( *pfMsgHandler )( const Msgs * ) )
    {
    Msgs msg;

    msg.msgId = MSG_ALLOCPLAYER;
    ( void ) RISK_SendSyncMessage( CLIENTS_GetCommLinkOfClient( CLNT_GetThisClientID( ) ),
                                   &msg,
                                   MSG_REPLYPACKET,
                                   pfMsgHandler );

    return ( iReply );
    }


/**
 * Free a player
 */
void CLNT_FreePlayer( Int32 i )
    {
    MsgFreePlayer mess;

    /* Tell the server that this player ID is free */
    mess.iPlayer = i;

    mess.msgId = MSG_FREEPLAYER;
    ( void ) RISK_SendMessage( CLIENTS_GetCommLinkOfClient( CLNT_GetThisClientID( ) ),
                               ( const Msgs * ) &mess );
    }


/**
 *
 */
Int32 CLNT_GetThisClientID( void )
    {
    return iCurrentClient;
    }

/**
 *
 */
Int32 CLNT_GetLightCountOfCountry( Int32 iCountry )
    {
    assert( iCountry >= 0 && iCountry < NUM_COUNTRIES );

    return piCountryLightCount[ iCountry ];
    }

/**
 *
 */
void CLNT_SetLightCountOfCountry( Int32 iCountry, Int32 iLightCount )
    {
    assert( iCountry >= 0 && iCountry < NUM_COUNTRIES );
    assert( iLightCount >= 0 );

    piCountryLightCount[ iCountry ] = iLightCount;
}

/**
 * Emergency exit function
 */
void FatalError( CString strError, Int32 iRetVal )
    {
    if ( strError )
        {
        printf( "%s\n", strError );
        }

    exit( iRetVal );
    }

/* End of file */
