/*
 *   Xfrisk - The classic board game for X
 *   Copyright (C) 1993-1999 Elan Feingold (elan@aetherworks.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   $Id: clients.c,v 1.20 2003/07/14 10:37:38 sketch Exp $
 */

/** \file
 * Client list.
 */

static const char
    rcsid[] = "@(#)$Id: clients.c,v 1.20 2003/07/14 10:37:38 sketch Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>

#include <assert.h>
#include <string.h>

#include "riskgame.h"
#include "clients.h"
#include "debug.h"
#include "language.h"


/* Private data types */

/**
 * Structure to track client failures
 */
typedef struct _Failure
    {
    CString     strReason;
    Int32       iCount;
    } Failure;


/**
 * Client data that doesn't get replicated to clients themselves.
 */
typedef struct _Client
    {
    Int32       iCommLink;
    CString     strAddress;
    Int32       iState;
    Int32       iType;
    Int32       iSpecies;
    Flag        fStarted;
    Failure     failure;
    } Client;


/* Private data structures */
static Client   pClients[ MAX_CLIENTS ];    /**< Client list */
static Int32    iNumClients  = 0;           /**< Number of clients actually on list */
static Int32    iNumFailures = 0;           /**< Total count of failures */


/**
 * Initialize client list.
 */
void CLIENTS_Init( void )
    {
    CLIENTS_SetNumClients( 0 );
    CLIENTS_SetNumFailures( 0 );
    memset( pClients, 0, sizeof( Client ) * MAX_CLIENTS );
    }


/**
 * Finish allocation of new client.
 */
Int32 CLIENTS_Alloc( void )
    {
    Int32 i;

    /* Since we wouldn't have accepted the connection if there weren't
     * a slot available, this call cannot fail.
     */
    assert( CLIENTS_GetNumClients( ) < MAX_CLIENTS );

    CLIENTS_SetNumClients( CLIENTS_GetNumClients( ) + 1 );

    /* Find an available client */
    for ( i = 0; i != MAX_CLIENTS; i++ )
        {
        if ( CLIENTS_GetAllocationState( i ) == ALLOC_NONE )
            {
            CLIENTS_SetAllocationState( i, ALLOC_COMPLETE );
            CLIENTS_SetCommLink( i, -1 );
            CLIENTS_SetStartState( i, FALSE );

            return ( i );
            }
        }

    assert( FALSE );

    /* For the compiler */
    return ( 0 );
    }


/**
 * Return failure count of given client.
 */
Int32 CLIENTS_GetFailureCount( Int32 iClient )
    {
    return pClients[ iClient ].failure.iCount;
    }


/**
 * Reset failure count of given client.
 */
void CLIENTS_ResetFailure( Int32 iNumClient )
    {
    pClients[ iNumClient ].failure.iCount = 0;
    pClients[ iNumClient ].failure.strReason = 0;
    }


/**
 * Set printable address of client
 *
 * \bug( MSH 25.12.99 ) Obsolete since client addresses are not disclosed to others anymore!
 */
void CLIENTS_SetAddress( Int32 iNumClient, CString strAddress )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    if ( pClients[ iNumClient ].strAddress != NULL )
        {
        MEM_Free( pClients[ iNumClient ].strAddress );
        }

    pClients[ iNumClient ].strAddress = MEM_Alloc( strlen( strAddress ) + 1 );
    strcpy( pClients[ iNumClient ].strAddress, strAddress );
    }


/**
 * Set allocation state of given client.
 */
void CLIENTS_SetAllocationState( Int32 iNumClient, Int32 iState )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    pClients[ iNumClient ].iState = iState;
    }


/**
 * Set communication socket of given client.
 */
void CLIENTS_SetCommLink( Int32 iNumClient, Int32 iCommLink )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    pClients[ iNumClient ].iCommLink = iCommLink;
    }


/**
 * Increase failure count and set failure reason string for given client.
 */
void CLIENTS_SetFailure( Int32 iClient, CString strReason )
    {
    pClients[ iClient ].failure.strReason = strReason;
    pClients[ iClient ].failure.iCount++;
    iNumFailures++;
    }

/**
 * Set client type.
 */
void CLIENTS_SetType( Int32 iNumClient, Int32 iType )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    pClients[ iNumClient ].iType = iType;
    }

/**
 * Set client species.
 */
void CLIENTS_SetSpecies( Int32 iNumClient, Int32 iSpecies )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    pClients[ iNumClient ].iSpecies = iSpecies;
    }

/**
 * Set number of clients.
 */
void CLIENTS_SetNumClients( Int32 iClients )
    {
    iNumClients = iClients;
    }

/**
 * Tell about other clients to given client.
 */
void CLIENTS_TellAboutOthers( Int32 iCommLink, Int32 iClientType )
    {
    char buf[ 256 ];
    int n;
    MsgMessagePacket      msgMess;

    if ( CLIENTS_GetNumClients( ) > 1 && iClientType != CLIENT_AI )
        {
        /* Let the new client know about old clients */
        msgMess.strMessage = buf;
        msgMess.iFrom      = FROM_SERVER;
        msgMess.iTo        = DST_OTHER;
        msgMess.msgId      = MSG_MESSAGEPACKET;

        /* Send info about everybody but the new client. */
        for ( n = 0; n != MAX_CLIENTS; n++ )
            {
            if ( CLIENTS_GetAllocationState( n ) == ALLOC_COMPLETE &&
                 CLIENTS_GetCommLinkOfClient( n ) != iCommLink )
                {
                snprintf( buf,
                          sizeof( buf ),
                          "%s %s.",
                          CLIENT,
                          IS_REGISTERED );

                ( void ) RISK_SendMessage( iCommLink, ( const Msgs * ) &msgMess );
                }
            }
        }
    }


/**
 * Return address of given client.
 */
CString CLIENTS_GetAddress( Int32 iNumClient )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    return pClients[ iNumClient ].strAddress;
    }


/**
 * Return type of given client.
 */
Int32 CLIENTS_GetType( Int32 iNumClient )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    return pClients[ iNumClient ].iType;
    }


/**
 * Return communication socket of given client.
 */
Int32 CLIENTS_GetCommLinkOfClient( Int32 iNumClient )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    return pClients[ iNumClient ].iCommLink;
    }


/**
 * Return allocation state of given client.
 */
Int32 CLIENTS_GetAllocationState( Int32 iNumClient )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    return pClients[ iNumClient ].iState;
    }


/**
 * Return species of given client.
 */
Int32 CLIENTS_GetSpecies( Int32 iNumClient )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    return pClients[ iNumClient ].iSpecies;
    }


/**
 * Return number of clients.
 */
Int32 CLIENTS_GetNumClients( void )
    {
    return iNumClients;
    }


/**
 * Return the client using given commucation socket.
 */
Int32 CLIENTS_GetCommLinkToClient( Int32 iCommLink )
    {
    Int32 i;

    assert( iCommLink >= 0 && iCommLink < MAX_DESCRIPTORS );

    for ( i = 0; i != MAX_CLIENTS; i++ )
        {
        if ( ( CLIENTS_GetAllocationState( i ) == ALLOC_COMPLETE ) &&
             ( CLIENTS_GetCommLinkOfClient( i ) == iCommLink ) )
            {
            return ( i );
            }
        }

    return ( -1 );
    }


/**
 * Return failure reason string for given client.
 */
CString CLIENTS_GetFailureReason( Int32 iNumClient )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    return ( pClients[ iNumClient ].failure.strReason );
    }


/**
 * Return failure count for given client.
 */
Int32 CLIENTS_GetNumFailures( void )
    {
    return iNumFailures;
    }

/**
 * Return number of clients that have finished registering players.
 */
Int32 CLIENTS_GetNumClientsStarted( void )
    {
    Int32 i, iCount;

    for ( i=iCount=0; i!=MAX_CLIENTS; i++ )
        {
        if ( CLIENTS_GetAllocationState( i ) == ALLOC_COMPLETE &&
             CLIENTS_GetStartState( i ) == TRUE )
            {
            iCount++;
            }
        }

    return iCount;
    }


/**
 * Return start state of given client.
 */
Int32 CLIENTS_GetStartState( Int32 iNumClient )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    return pClients[ iNumClient ].fStarted;
    }


/**
 * Set start state of given client.
 */
void CLIENTS_SetStartState( Int32 iNumClient, Int32 iState )
    {
    assert( iNumClient >= 0 && iNumClient < MAX_CLIENTS );

    pClients[ iNumClient ].fStarted = iState;
    }


/**
 * Set number of failures for given client.
 */
void CLIENTS_SetNumFailures( Int32 num )
    {
    iNumFailures = num;
    }

/* EOF */
