/*
 *   Xfrisk - The classic board game for X
 *   Copyright (C) 1993-1999 Elan Feingold (elan@aetherworks.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   $Id: deck.c,v 1.25 2003/07/14 10:37:38 sketch Exp $
 */

/** \file
 * Card deck handling routines for the server
 */

static const char
    rcsid[] = "@(#)$Id: deck.c,v 1.25 2003/07/14 10:37:38 sketch Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>

#include "deck.h"
#include "debug.h"
#include "riskgame.h"

/**
 * Structure for card deck data.
 */
struct _Deck
    {
    Int32   iCardsLeft;         /**< Number of cards still in the deck */
    Int32   iCardsReturned;     /**< Number of cards returned to the deck */
    Int32   iTotalCards;        /**< Total number of cards in deck */
    Int32  *piCards;            /**< Pointer to list of cards still in the deck */
    Int32  *piCardsReturned;    /**< Pointer to list of cards returned to the deck */
    };


/**
 * Create a card deck.
 *
 * \param iNumCards Number of cards to be put in the deck.
 */
Deck *DECK_Create( Int32 iNumCards )
    {
    Deck   *pDeck;
    Int32   i;

    pDeck = ( Deck * ) MEM_Alloc( sizeof( Deck ) );

    /* Init the structure */
    pDeck->iCardsLeft       = pDeck->iTotalCards = iNumCards;
    pDeck->iCardsReturned   = 0;
    pDeck->piCards          = ( Int32 * ) MEM_Alloc( sizeof( Int32 ) * iNumCards );
    pDeck->piCardsReturned  = ( Int32 * ) MEM_Alloc( sizeof( Int32 ) * iNumCards );

    /* Init all of the cards */
    for ( i = 0; i != iNumCards; i++ )
        {
        pDeck->piCards[ i ] = i;
        }

    return ( pDeck );
    }


/**
 * Draw a random card from the deck.
 *
 * \param pDeck Pointer to the deck
 */
Int32 DECK_GetCard( Deck *pDeck )
    {
    Int32   iCardIndex;
    Int32   iCard;

    /* If there are no cards left, take all of the cards from the
     * returned pile and put them into the deck.
     */

    if ( !pDeck->iCardsLeft )
        {
        /* Are there cards to return to the deck? */
        if ( !pDeck->iCardsReturned )
            {
            return ( -1 );
            }

        /* Copy all of the returned cards to the deck */
        memcpy( ( char * ) pDeck->piCards,
                ( char * ) pDeck->piCardsReturned,
                sizeof( Int32 ) * pDeck->iCardsReturned );
        pDeck->iCardsLeft       = pDeck->iCardsReturned;
        pDeck->iCardsReturned   = 0;

        }

    /* Pick a card out of the remaining ones */
    iCardIndex = rand( ) % pDeck->iCardsLeft;

    /* Switch the last card with this one. */
    iCard = pDeck->piCards[ iCardIndex ];
    pDeck->piCards[ iCardIndex ] = pDeck->piCards[ --pDeck->iCardsLeft ];

    return ( iCard );
    }


/**
 * Add( return ) a card to the deck.
 *
 * \param pDeck Pointer to the deck
 * \param iCard Number of card to add
 */
void DECK_PutCard( Deck *pDeck, Int32 iCard )
    {
    /* If returned card pile fills up, this is bad */
    assert( pDeck );

    if ( ( pDeck->iCardsLeft + pDeck->iCardsReturned ) >= pDeck->iTotalCards )
        {
        printf( "Fatal Error! DECK: Can't add to full deck!\n" );
        return;
        }

    /* Add the card to the retured cards pile */
    pDeck->piCardsReturned[ pDeck->iCardsReturned++ ] = iCard;
    }


/**
 * Destroy the deck.
 */
void DECK_Destroy( Deck *pDeck )
    {
    MEM_Free( pDeck->piCards );
    MEM_Free( pDeck->piCardsReturned );
    MEM_Free( pDeck );
    }

/**
 * Exchange 3 given cards for armies
 *
 * \param pDeck Pointer to the card deck
 * \param piCards Pointer to indices of 3 cards to be exchanged
 */
Int32 DECK_ExchangeCards( Deck *pDeck, Int32 *piCards )
    {
    Int32   iNumTypes[ NUM_CARDTYPES ] = { 0 };
    Int32   n;
    Int32   retval = 12;

    assert( pDeck );
    assert( piCards );

    /* Put cards back on the deck and change them to type */
    for ( n = 0; n < 3; n++ )
        {
        DECK_PutCard( pDeck, piCards[ n ] );

        /*
         * Crude hack; map every 3rd country to each card type
         */
        if ( piCards[ n ] < NUM_COUNTRIES )
            {
            iNumTypes[ piCards[ n ] % 3 ]++;
            assert( iNumTypes[ piCards[ n ] % 3 ] > 0 );
            assert( iNumTypes[ piCards[ n ] % 3 ] < 4 );
            }
        else
            {
            /*
             * Cards at and above NUM_COUNTRIES are jokers
             */
            iNumTypes[ CARD_JOKER ]++;
            assert( iNumTypes[ CARD_JOKER ] > 0 );
            assert( iNumTypes[ CARD_JOKER ] < 4 );
            }
        }

    /* Find out how many armies the player gets in
     * exchange for the cards
     */

    if ( ( iNumTypes[ CARD_CAVALRY ] == 3 ) ||
         ( ( iNumTypes[ CARD_CAVALRY ] == 2 ) &&
           ( iNumTypes[ CARD_JOKER ] == 1 ) ) )
        {
        retval = 8;
        }
    else if ( ( iNumTypes[ CARD_INFANTRY ] == 3 ) ||
              ( ( iNumTypes[ CARD_INFANTRY ] == 2 ) &&
                ( iNumTypes[ CARD_JOKER ] == 1 ) ) )
        {
        retval = 6;
        }
    else if ( ( iNumTypes[ CARD_ARTILLERY ] == 3 ) ||
              ( ( iNumTypes[ CARD_ARTILLERY ] == 2 ) &&
                ( iNumTypes[ CARD_JOKER ] == 1 ) ) )
        {
        retval = 4;
        }
    else if ( ( iNumTypes[ CARD_JOKER ] >= 2 ) ||
              ( ( iNumTypes[ CARD_JOKER ]     < 2 ) &&
                ( iNumTypes[ CARD_CAVALRY ]   < 2 ) &&
                ( iNumTypes[ CARD_INFANTRY ]  < 2 ) &&
                ( iNumTypes[ CARD_ARTILLERY ] < 2 ) ) )
        {
        /*
         * 3 jokers, 2 jokers and any card or any three cards of different
         * types
         */
        retval = 10;
        }
    else
        {
        /*
         * Special case to trap errors
         */
        assert( !"Army determination code for card exchange has failed" );
        }

    return ( retval );

    }

/* End of file */
