/*
 *   Xfrisk - The classic board game for X
 *   Copyright (C) 1993-1999 Elan Feingold (elan@aetherworks.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   $Id: network.c,v 1.22 2003/07/14 10:37:38 sketch Exp $
 */

/** \file
 * Messaging interface to/from the network.
 */

static const char
    rcsid[] = "@(#)$Id: network.c,v 1.22 2003/07/14 10:37:38 sketch Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <unistd.h>

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include "types.h"
#include "riskgame.h"
#include "network.h"
#include "debug.h"

/* Local data types */
typedef Int32 ( *NET_SendFuncPtr )( Int32 iSocket, const Msgs *pMsg );
typedef Int32 ( *NET_RecvFuncPtr )( Int32 iSocket, Msgs *pMsg );

typedef struct
    {
    MsgId           msgId;
    NET_SendFuncPtr fPtr;
    } NET_SendEntry;

typedef struct
    {
    MsgId           msgId;
    NET_RecvFuncPtr fPtr;
    } NET_RecvEntry;

/* Local prototypes */
static Int32 NET_SendString( Int32 iSocket, CString strCString );
static Int32 NET_SendLong( Int32 iSocket, Int32 iLong );
static Int32 NET_RecvString( Int32 iSocket, CString *pstrCString );
static Int32 NET_SocketRead( Int32 iSocket, void *ptr, Int32 iNumBytes );
static Int32 NET_SocketWrite( Int32 iSocket, void *ptr, Int32 iNumBytes );
static Int32 NET_RecvLong( Int32 iSocket, Int32 *piLong );

/* Prototypes for message-specific send functions */
static Int32 NET_S_MsgOldRegisterClient( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_ExchangeCards( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_RequestCard( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_CardPacket( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_ReplyPacket( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_SendMessage( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_MessagePacket( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_Exit( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_StartGame( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_TurnNotify( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_EndTurn( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_ClientIdent( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_EnterState( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_EndOfGame( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_ObjStrUpdate( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_ObjIntUpdate( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_DeRegisterClient( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_AllocPlayer( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_FreePlayer( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_NetMessage( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_NetPopup( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_PopupRegisterBox( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_DiceRoll( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_PlaceNotify( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_AttackNotify( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_MoveNotify( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_PollClients( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_Hello( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_Version( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_SpeciesIdent( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_Mission( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_EndOfMission( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_Victory( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_ForceExchangeCards( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_RegisterClient( Int32 iSocket, const Msgs *pMsg );
static Int32 NET_S_NoMessage( Int32 iSocket, const Msgs *pMsg );

/* Prototypes for message-specific receive functions */
static Int32 NET_R_MsgOldRegisterClient( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_ExchangeCards( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_RequestCard( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_CardPacket( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_ReplyPacket( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_SendMessage( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_MessagePacket( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_Exit( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_StartGame( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_TurnNotify( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_EndTurn( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_ClientIdent( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_EnterState( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_EndOfGame( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_ObjStrUpdate( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_ObjIntUpdate( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_DeRegisterClient( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_AllocPlayer( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_FreePlayer( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_NetMessage( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_NetPopup( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_PopupRegisterBox( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_DiceRoll( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_PlaceNotify( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_AttackNotify( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_MoveNotify( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_PollClients( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_Hello( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_Version( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_SpeciesIdent( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_Mission( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_EndOfMission( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_Victory( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_ForceExchangeCards( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_RegisterClient( Int32 iSocket, Msgs *pMsg );
static Int32 NET_R_NoMessage( Int32 iSocket, Msgs *pMsg );

/* Local data structures */
static NET_SendEntry pSendFunctions[ ] =
    {
    { MSG_UNUSED_1,             NULL },
    { MSG_OLDREGISTERCLIENT,    NET_S_MsgOldRegisterClient },
    { MSG_UNUSED_2,             NULL },
    { MSG_EXCHANGECARDS,        NET_S_ExchangeCards },
    { MSG_REQUESTCARD,          NET_S_RequestCard },
    { MSG_CARDPACKET,           NET_S_CardPacket },
    { MSG_REPLYPACKET,          NET_S_ReplyPacket },
    { MSG_SENDMESSAGE,          NET_S_SendMessage },
    { MSG_MESSAGEPACKET,        NET_S_MessagePacket },
    { MSG_EXIT,                 NET_S_Exit },
    { MSG_STARTGAME,            NET_S_StartGame },
    { MSG_TURNNOTIFY,           NET_S_TurnNotify },
    { MSG_ENDTURN,              NET_S_EndTurn },
    { MSG_CLIENTIDENT,          NET_S_ClientIdent },
    { MSG_ENTERSTATE,           NET_S_EnterState },
    { MSG_UNUSED_3,             NULL },
    { MSG_ENDOFGAME,            NET_S_EndOfGame },
    { MSG_OBJSTRUPDATE,         NET_S_ObjStrUpdate },
    { MSG_OBJINTUPDATE,         NET_S_ObjIntUpdate },
    { MSG_UNUSED_4,             NULL },
    { MSG_UNUSED_5,             NULL },
    { MSG_UNUSED_6,             NULL },
    { MSG_DEREGISTERCLIENT,     NET_S_DeRegisterClient },
    { MSG_UNUSED_7,             NULL },
    { MSG_UNUSED_8,             NULL },
    { MSG_ALLOCPLAYER,          NET_S_AllocPlayer },
    { MSG_UNUSED_9,             NULL },
    { MSG_UNUSED_10,            NULL },
    { MSG_UNUSED_11,            NULL },
    { MSG_UNUSED_12,            NULL },
    { MSG_UNUSED_13,            NULL },
    { MSG_UNUSED_14,            NULL },
    { MSG_FREEPLAYER,           NET_S_FreePlayer },
    { MSG_NETMESSAGE,           NET_S_NetMessage },
    { MSG_NETPOPUP,             NET_S_NetPopup },
    { MSG_POPUPREGISTERBOX,     NET_S_PopupRegisterBox },
    { MSG_DICEROLL,             NET_S_DiceRoll },
    { MSG_PLACENOTIFY,          NET_S_PlaceNotify },
    { MSG_ATTACKNOTIFY,         NET_S_AttackNotify },
    { MSG_MOVENOTIFY,           NET_S_MoveNotify },
    { MSG_POLLCLIENTS,          NET_S_PollClients },
    { MSG_UNUSED_15,            NULL },
    { MSG_HELLO,                NET_S_Hello },
    { MSG_VERSION,              NET_S_Version },
    { MSG_SPECIESIDENT,         NET_S_SpeciesIdent },
    { MSG_MISSION,              NET_S_Mission },
    { MSG_ENDOFMISSION,         NET_S_EndOfMission },
    { MSG_VICTORY,              NET_S_Victory },
    { MSG_FORCEEXCHANGECARDS,   NET_S_ForceExchangeCards },
    { MSG_NUM_CONTIGUOUS,       NULL },
    { MSG_REGISTERCLIENT,       NET_S_RegisterClient },
    { MSG_NOMESSAGE,            NET_S_NoMessage },
    { MSG_MAX_ID,               NULL }
    };

static NET_RecvEntry pRecvFunctions[ ] =
    {
    { MSG_UNUSED_1,             NULL },
    { MSG_OLDREGISTERCLIENT,    NET_R_MsgOldRegisterClient },
    { MSG_UNUSED_2,             NULL },
    { MSG_EXCHANGECARDS,        NET_R_ExchangeCards },
    { MSG_REQUESTCARD,          NET_R_RequestCard },
    { MSG_CARDPACKET,           NET_R_CardPacket },
    { MSG_REPLYPACKET,          NET_R_ReplyPacket },
    { MSG_SENDMESSAGE,          NET_R_SendMessage },
    { MSG_MESSAGEPACKET,        NET_R_MessagePacket },
    { MSG_EXIT,                 NET_R_Exit },
    { MSG_STARTGAME,            NET_R_StartGame },
    { MSG_TURNNOTIFY,           NET_R_TurnNotify },
    { MSG_ENDTURN,              NET_R_EndTurn },
    { MSG_CLIENTIDENT,          NET_R_ClientIdent },
    { MSG_ENTERSTATE,           NET_R_EnterState },
    { MSG_UNUSED_3,             NULL },
    { MSG_ENDOFGAME,            NET_R_EndOfGame },
    { MSG_OBJSTRUPDATE,         NET_R_ObjStrUpdate },
    { MSG_OBJINTUPDATE,         NET_R_ObjIntUpdate },
    { MSG_UNUSED_4,             NULL },
    { MSG_UNUSED_5,             NULL },
    { MSG_UNUSED_6,             NULL },
    { MSG_DEREGISTERCLIENT,     NET_R_DeRegisterClient },
    { MSG_UNUSED_7,             NULL },
    { MSG_UNUSED_8,             NULL },
    { MSG_ALLOCPLAYER,          NET_R_AllocPlayer },
    { MSG_UNUSED_9,             NULL },
    { MSG_UNUSED_10,            NULL },
    { MSG_UNUSED_11,            NULL },
    { MSG_UNUSED_12,            NULL },
    { MSG_UNUSED_13,            NULL },
    { MSG_UNUSED_14,            NULL },
    { MSG_FREEPLAYER,           NET_R_FreePlayer },
    { MSG_NETMESSAGE,           NET_R_NetMessage },
    { MSG_NETPOPUP,             NET_R_NetPopup },
    { MSG_POPUPREGISTERBOX,     NET_R_PopupRegisterBox },
    { MSG_DICEROLL,             NET_R_DiceRoll },
    { MSG_PLACENOTIFY,          NET_R_PlaceNotify },
    { MSG_ATTACKNOTIFY,         NET_R_AttackNotify },
    { MSG_MOVENOTIFY,           NET_R_MoveNotify },
    { MSG_POLLCLIENTS,          NET_R_PollClients },
    { MSG_UNUSED_15,            NULL },
    { MSG_HELLO,                NET_R_Hello },
    { MSG_VERSION,              NET_R_Version },
    { MSG_SPECIESIDENT,         NET_R_SpeciesIdent },
    { MSG_MISSION,              NET_R_Mission },
    { MSG_ENDOFMISSION,         NET_R_EndOfMission },
    { MSG_VICTORY,              NET_R_Victory },
    { MSG_FORCEEXCHANGECARDS,   NET_R_ForceExchangeCards },
    { MSG_NUM_CONTIGUOUS,       NULL },
    { MSG_REGISTERCLIENT,       NET_R_RegisterClient },
    { MSG_NOMESSAGE,            NET_R_NoMessage },
    { MSG_MAX_ID,               NULL }
    };

/* Private to this routine */
static char strLogging[ 1024 ];

 /**
 * Send a message to the given socket
 */
Int32 NET_SendMessage( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = -1;
    Flag found = FALSE;
    Int32 i;

    /* Sanity check */
    if ( pMsg == NULL )
        {
        return ( retval );
        }

    /* Find sender function */
    if ( pMsg->msgId < MSG_NUM_CONTIGUOUS )
        {
        i = pMsg->msgId;
        found = TRUE;
        }
    else
        {
        i = ( Int32 ) MSG_NUM_CONTIGUOUS;
        while ( pSendFunctions[ i ].msgId != MSG_MAX_ID )
            {
            if ( ( pSendFunctions[ i ].msgId == pMsg->msgId ) &&
                 ( pSendFunctions[ i ].fPtr != NULL ) )
                {
                found = TRUE;
                break;
                }
            i++;
            }
        }

    /* Send the message if possible */
    if ( found && pSendFunctions[ i ].fPtr != NULL )
        {
#ifndef NDEBUG
        printf( "Sending message ID %.2x\n", pMsg->msgId );
#endif /* NDEBUG */
        retval = ( pSendFunctions[ i ].fPtr )( iSocket, pMsg );
#ifndef NDEBUG
        printf( "Sending message ID %.2x complete, status %i\n",
                pMsg->msgId,
                retval );
#endif /* NDEBUG */
        }
    else if ( !found )
        {
#ifndef NDEBUG
        printf( "NETWORK ERROR: Sending of unknown message ID %.2x attempted\n", pMsg->msgId );
#endif /* NDEBUG */
        retval = -1;
        }
    else /* found && fPtr == NULL */
        {
#ifndef NDEBUG
        printf( "NETWORK ERROR: Sending of unused message ID %.2x attempted\n",
                pMsg->msgId );
#endif /* NDEBUG */
        retval = -1;
        }

    return ( retval );
    }

/*
 * Message-specific send functions
 */

static Int32 NET_S_MsgOldRegisterClient( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval |= NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendString( iSocket,
                              pMsg->msgOldRegisterClient.strClientAddress );

    return ( retval );
    }

static Int32 NET_S_ExchangeCards( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket,
                           pMsg->msgId );
    retval |= NET_SendLong( iSocket,
                            pMsg->msgExchangeCards.piCards[ 0 ] );
    retval |= NET_SendLong( iSocket,
                            pMsg->msgExchangeCards.piCards[ 1 ] );
    retval |= NET_SendLong( iSocket,
                            pMsg->msgExchangeCards.piCards[ 2 ] );

    return ( retval );
    }

static Int32 NET_S_RequestCard( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgRequestCard.iPlayer );

    return ( retval );
    }

static Int32 NET_S_CardPacket( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, ( Int32 ) pMsg->msgCardPacket.iPlayer );
    retval |= NET_SendLong( iSocket, ( Int32 ) pMsg->msgCardPacket.cdCard );

    return ( retval );
    }

static Int32 NET_S_ReplyPacket( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, ( Int32 ) pMsg->msgReplyPacket.iReply );

    return ( retval );
    }

static Int32 NET_S_SendMessage( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendString( iSocket, pMsg->msgSendMessage.strMessage );
    retval |= NET_SendString( iSocket, pMsg->msgSendMessage.strDestination );

    return ( retval );
    }

static Int32 NET_S_MessagePacket( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendString( iSocket, pMsg->msgMessagePacket.strMessage );
    retval |= NET_SendLong( iSocket, pMsg->msgMessagePacket.iFrom );
    retval |= NET_SendLong( iSocket, pMsg->msgMessagePacket.iTo );

    return ( retval );
    }

static Int32 NET_S_Exit( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    return ( retval );
    }

static Int32 NET_S_StartGame( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    return ( retval );
    }

static Int32 NET_S_TurnNotify( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgTurnNotify.iPlayer );
    retval |= NET_SendLong( iSocket, pMsg->msgTurnNotify.iClient );

    return ( retval );
    }

static Int32 NET_S_EndTurn( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    return ( retval );
    }

static Int32 NET_S_ClientIdent( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgClientIdent.iClientID );

    return ( retval );
    }

static Int32 NET_S_EnterState( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgEnterState.iState );

    return ( retval );
    }

static Int32 NET_S_EndOfGame( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    return ( retval );
    }

static Int32 NET_S_ObjStrUpdate( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgObjStrUpdate.iField );
    retval |= NET_SendLong( iSocket, pMsg->msgObjStrUpdate.iIndex1 );
    retval |= NET_SendLong( iSocket, pMsg->msgObjStrUpdate.iIndex2 );
    retval |= NET_SendString( iSocket, pMsg->msgObjStrUpdate.strNewValue );

    return ( retval );
    }

static Int32 NET_S_ObjIntUpdate( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgObjIntUpdate.iField );
    retval |= NET_SendLong( iSocket, pMsg->msgObjIntUpdate.iIndex1 );
    retval |= NET_SendLong( iSocket, pMsg->msgObjIntUpdate.iIndex2 );
    retval |= NET_SendLong( iSocket, pMsg->msgObjIntUpdate.iNewValue );

    return ( retval );
    }

static Int32 NET_S_DeRegisterClient( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    return ( retval );
    }

static Int32 NET_S_AllocPlayer( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    return ( retval );
    }

static Int32 NET_S_FreePlayer( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgFreePlayer.iPlayer );

    return ( retval );
    }

static Int32 NET_S_NetMessage( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendString( iSocket, pMsg->msgNetMessage.strMessage );

    return ( retval );
    }

static Int32 NET_S_NetPopup( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendString( iSocket, pMsg->msgNetPopup.strMessage );

    return ( retval );
    }

static Int32 NET_S_PopupRegisterBox( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    return ( retval );
    }

static Int32 NET_S_DiceRoll( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;
    Int32 i;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    for ( i = 0; i != 3; i++ )
        {
        retval |= NET_SendLong( iSocket,
                                pMsg->msgDiceRoll.pAttackDice[ i ] );
        }

    for ( i = 0; i != 3; i++ )
        {
        retval |= NET_SendLong( iSocket,
                                pMsg->msgDiceRoll.pDefendDice[ i ] );
        }

    retval |= NET_SendLong( iSocket,
                            pMsg->msgDiceRoll.iDefendingPlayer );

    return ( retval );
    }

static Int32 NET_S_PlaceNotify( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgPlaceNotify.iCountry );

    return ( retval );
    }

static Int32 NET_S_AttackNotify( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgAttackNotify.iSrcCountry );
    retval |= NET_SendLong( iSocket, pMsg->msgAttackNotify.iDstCountry );

    return ( retval );
    }

static Int32 NET_S_MoveNotify( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgMoveNotify.iSrcCountry );
    retval |= NET_SendLong( iSocket, pMsg->msgMoveNotify.iDstCountry );

    return ( retval );
    }

static Int32 NET_S_PollClients( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    return ( retval );
    }

static Int32 NET_S_Hello( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    return ( retval );
    }

static Int32 NET_S_Version( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendString( iSocket, pMsg->msgVersion.strVersion );

    return ( retval );
    }

static Int32 NET_S_SpeciesIdent( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgSpeciesIdent.iSpeciesID );

    return ( retval );
    }

static Int32 NET_S_Mission( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    return ( retval );
    }

static Int32 NET_S_EndOfMission( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgEndOfMission.iWinner );
    retval |= NET_SendLong( iSocket, pMsg->msgEndOfMission.iTyp );
    retval |= NET_SendLong( iSocket, pMsg->msgEndOfMission.iNum1 );
    retval |= NET_SendLong( iSocket, pMsg->msgEndOfMission.iNum2 );

    return ( retval );
    }

static Int32 NET_S_Victory( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgVictory.iWinner );

    return ( retval );
    }

static Int32 NET_S_ForceExchangeCards( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendLong( iSocket, pMsg->msgForceExchangeCards.iPlayer );

    return ( retval );
    }

static Int32 NET_S_RegisterClient( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );
    retval |= NET_SendString( iSocket,
                              pMsg->msgRegisterClient.strClientAddress );
    retval |= NET_SendLong( iSocket,
                            pMsg->msgRegisterClient.iClientType );

    return ( retval );
    }

static Int32 NET_S_NoMessage( Int32 iSocket, const Msgs *pMsg )
    {
    Int32 retval = 0;

    retval = NET_SendLong( iSocket, ( Int32 ) pMsg->msgId );

    return ( retval );
    }

/**
 * Receive one message from the network
 */
Int32 NET_RecvMessage( Int32 iSocket, Msgs **ppMsg )
    {
    Int32   retval = -1;
    Int32   i;
    Flag    found = FALSE;
    Msgs   *pMsg;

    assert( ppMsg );

    pMsg = MEM_Alloc( sizeof( Msgs ) );

    /* Get the message ID */
    retval = NET_RecvLong( iSocket, ( Int32 * ) &( pMsg->msgId ) );

    if ( retval == 0 )
        {
#ifndef NDEBUG
        printf( "NETWORK: Looking for receiver function for message ID %.2x, status %i\n",
                pMsg->msgId,
                retval );
#endif /* NDEBUG */
        /* Find receiver function */
        if ( pMsg->msgId < MSG_NUM_CONTIGUOUS )
            {
            i = pMsg->msgId;
            found = TRUE;
            }
        else
            {
            i = ( Int32 ) MSG_NUM_CONTIGUOUS;
            while ( pRecvFunctions[ i ].msgId != MSG_MAX_ID )
                {
                if ( ( pRecvFunctions[ i ].msgId == pMsg->msgId ) &&
                     ( pRecvFunctions[ i ].fPtr != NULL ) )
                    {
                    found = TRUE;
                    break;
                    }
                i++;
                }
            }
        }
    else
        {
#ifndef NDEBUG
        printf( "NETWORK ERROR: read failed - %s\n", strerror( errno ) );
#endif /* NDEBUG */
        }

    if ( found && pRecvFunctions[ i ].fPtr != NULL )
        {
#ifndef NDEBUG
        printf( "Receiving message ID %.2x\n", pMsg->msgId );
#endif /* NDEBUG */
        retval = ( pRecvFunctions[ i ].fPtr )( iSocket, pMsg );
#ifndef NDEBUG
        printf( "Receiving message ID %.2x complete, status %i\n",
                pMsg->msgId,
                retval );
#endif /* NDEBUG */
        }
    else
        {
#ifndef NDEBUG
        printf( "NETWORK ERROR: Received unknown message ID %.2x\n", pMsg->msgId );
#endif /* NDEBUG */
        retval = -1;
        }

    if ( retval )
        {
        MEM_Free( pMsg );
        *ppMsg = NULL;
        }
    else
        {
        *ppMsg = pMsg;
        }

    return ( retval );
    }

/*
 * Message-specific receive functions
 */

static Int32 NET_R_MsgOldRegisterClient( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval = NET_RecvString( iSocket,
                             &( pMsg->msgOldRegisterClient.strClientAddress ) );

    return ( retval );
    }

static Int32 NET_R_ExchangeCards( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    Int32   i;

    for( i = 0; i != 3; i++ )
        {
        retval |= NET_RecvLong( iSocket,
                                ( Int32 * ) &( pMsg->msgExchangeCards.piCards[ i ] ) );
        }

    return ( retval );
    }

static Int32 NET_R_RequestCard( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval = NET_RecvLong( iSocket,
                           &( pMsg->msgRequestCard.iPlayer ) );

    return ( retval );
    }

static Int32 NET_R_CardPacket( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval = NET_RecvLong( iSocket,
                           ( Int32 * ) &( pMsg->msgCardPacket.iPlayer ) );
    retval |= NET_RecvLong( iSocket,
                           ( Int32 * ) &( pMsg->msgCardPacket.cdCard ) );

    return ( retval );
    }

static Int32 NET_R_ReplyPacket( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval = NET_RecvLong( iSocket,
                           ( Int32 * ) &( pMsg->msgReplyPacket.iReply ) );

    return ( retval );
    }

static Int32 NET_R_SendMessage( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval = NET_RecvString( iSocket,
                             &( pMsg->msgSendMessage.strMessage ) );
    retval |= NET_RecvString( iSocket,
                              &( pMsg->msgSendMessage.strDestination ) );

    return ( retval );
    }

static Int32 NET_R_MessagePacket( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval = NET_RecvString( iSocket,
                             &( pMsg->msgMessagePacket.strMessage ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgMessagePacket.iFrom ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgMessagePacket.iTo ) );

    return ( retval );
    }

static Int32 NET_R_Exit( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    UNUSED( iSocket );
    UNUSED( pMsg );

    return ( retval );
    }

static Int32 NET_R_StartGame( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    UNUSED( iSocket );
    UNUSED( pMsg );

    return ( retval );
    }

static Int32 NET_R_TurnNotify( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgTurnNotify.iPlayer ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgTurnNotify.iClient ) );

    return ( retval );
    }

static Int32 NET_R_EndTurn( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    UNUSED( iSocket );
    UNUSED( pMsg );

    return ( retval );
    }

static Int32 NET_R_ClientIdent( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgClientIdent.iClientID ) );

    return ( retval );
    }

static Int32 NET_R_EnterState( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgEnterState.iState ) );

    return ( retval );
    }

static Int32 NET_R_EndOfGame( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    UNUSED( iSocket );
    UNUSED( pMsg );

    return ( retval );
    }

static Int32 NET_R_ObjStrUpdate( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgObjStrUpdate.iField ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgObjStrUpdate.iIndex1 ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgObjStrUpdate.iIndex2 ) );
    retval |= NET_RecvString( iSocket,
                              &( pMsg->msgObjStrUpdate.strNewValue ) );

    return ( retval );
    }

static Int32 NET_R_ObjIntUpdate( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgObjIntUpdate.iField ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgObjIntUpdate.iIndex1 ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgObjIntUpdate.iIndex2 ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgObjIntUpdate.iNewValue ) );

    return ( retval );
    }

static Int32 NET_R_DeRegisterClient( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    UNUSED( iSocket );
    UNUSED( pMsg );

    return ( retval );
    }

static Int32 NET_R_AllocPlayer( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    UNUSED( iSocket );
    UNUSED( pMsg );

    return ( retval );
    }

static Int32 NET_R_FreePlayer( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgFreePlayer.iPlayer ) );

    return ( retval );
    }

static Int32 NET_R_NetMessage( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvString( iSocket,
                              &( pMsg->msgNetMessage.strMessage ) );

    return ( retval );
    }

static Int32 NET_R_NetPopup( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvString( iSocket,
                              &( pMsg->msgNetPopup.strMessage ) );

    return ( retval );
    }

static Int32 NET_R_PopupRegisterBox( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    UNUSED( iSocket );
    UNUSED( pMsg );

    return ( retval );
    }

static Int32 NET_R_DiceRoll( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    Int32   i;

    for ( i = 0; i != 3; i++ ) 
        {
        retval |= NET_RecvLong( iSocket,
                                &( pMsg->msgDiceRoll.pAttackDice[ i ] ) );
        }

    for ( i = 0; i != 3; i++ )
        {
        retval |= NET_RecvLong( iSocket,
                                &( pMsg->msgDiceRoll.pDefendDice[ i ] ) );
        }

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgDiceRoll.iDefendingPlayer ) );

    return ( retval );
    }

static Int32 NET_R_PlaceNotify( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgPlaceNotify.iCountry ) );

    return ( retval );
    }

static Int32 NET_R_AttackNotify( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgAttackNotify.iSrcCountry ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgAttackNotify.iDstCountry ) );

    return ( retval );
    }

static Int32 NET_R_MoveNotify( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgMoveNotify.iSrcCountry ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgMoveNotify.iDstCountry ) );

    return ( retval );
    }

static Int32 NET_R_PollClients( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    UNUSED( iSocket );
    UNUSED( pMsg );

    return ( retval );
    }

static Int32 NET_R_Hello( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    UNUSED( iSocket );
    UNUSED( pMsg );

    return ( retval );
    }

static Int32 NET_R_Version( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvString( iSocket,
                              &( pMsg->msgVersion.strVersion ) );

    return ( retval );
    }

static Int32 NET_R_SpeciesIdent( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgSpeciesIdent.iSpeciesID ) );

    return ( retval );
    }

static Int32 NET_R_Mission( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    UNUSED( iSocket );
    UNUSED( pMsg );

    return ( retval );
    }

static Int32 NET_R_EndOfMission( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgEndOfMission.iWinner ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgEndOfMission.iTyp ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgEndOfMission.iNum1 ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgEndOfMission.iNum2 ) );

    return ( retval );
    }

static Int32 NET_R_Victory( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgVictory.iWinner ) );

    return ( retval );
    }

static Int32 NET_R_ForceExchangeCards( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgForceExchangeCards.iPlayer ) );

    return ( retval );
    }

static Int32 NET_R_RegisterClient( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;

    retval |= NET_RecvString( iSocket,
                              &( pMsg->msgRegisterClient.strClientAddress ) );
    retval |= NET_RecvLong( iSocket,
                            &( pMsg->msgRegisterClient.iClientType ) );

    return ( retval );
    }

static Int32 NET_R_NoMessage( Int32 iSocket, Msgs *pMsg )
    {
    Int32   retval = 0;
    UNUSED( iSocket );
    UNUSED( pMsg );

    return ( retval );
    }

/**
 * Send a C-style nul-terminated string to the given socket
 */
static Int32 NET_SendString( Int32 iSocket, CString strCString )
    {
    Int32 iLength;
    Int32 retval = -1;

    /* Send the length and then the string itself */
    if ( strCString == NULL ) 
        {
        iLength = 0;
        }
    else
        {
        iLength = strlen( strCString ) + 1;
        }

    if ( NET_SendLong( iSocket, ( Int32 ) iLength ) == 0 )
        {
        if ( NET_SocketWrite( iSocket, ( Char * ) strCString, ( Int32 ) iLength ) ==
             iLength )
            {
            retval = 0;
            }
        }

#ifndef NDEBUG
    printf( "NET_SendString: string \"%s\", status %i\n", strCString, retval );
#endif /* NDEBUG */
    return ( retval );
    }


/**
 * Send a 32bit integer to the given socket
 */
static Int32 NET_SendLong( Int32 iSocket, Int32 iLong ) 
    {
    Int32 retval = 0;
    Int32 iData;

    iData = htonl( iLong );

    if ( NET_SocketWrite( iSocket, &iData, sizeof( Int32 ) ) != sizeof( Int32 ) )
        {
        retval = -1;
        }

    return ( retval );
    }


/**
 * Receive a string from the given socket. First reads the length
 * (32bit value) , then the string data.
 *
 * \bug Assumes that a nul terminating byte is in the data stream
 */
static Int32 NET_RecvString( Int32 iSocket, CString *pstrCString ) 
    {
    Int32       iRet;
    Int32       retval = 0;
    Int32       iLength;
    CString     strTemp;

    /* Receive the length and then the byte stream */
    NET_RecvLong( iSocket, &iLength );

    if ( iLength ) 
        {
        strTemp = ( CString ) MEM_Alloc( iLength );
        iRet = NET_SocketRead( iSocket, strTemp, iLength );

        /* Return an error if less than the string was read! */
        if ( ( iRet < 0 ) || ( iRet != iLength ) )
            {
            retval = -1;
            }
        }
    else
        {
        retval = -1;
        strTemp = NULL;
        }

    *pstrCString = strTemp;

    return ( retval );
    }


/**
 * Reads a 32bit value from the given socket.
 */
static Int32 NET_RecvLong( Int32 iSocket, Int32 *piLong ) 
    {
    Int32 iRet;
    Int32 retval = 0;

    iRet = NET_SocketRead( iSocket, piLong, sizeof( Int32 ) );

    /* Adjust for the network munging */
    *piLong = ntohl( *piLong );

    /* Return an error if read returns an error or if the wrong number
     * of bytes come across -- there should be sizeof( Int32 ) bytes!
     */

    if ( ( iRet < 0 ) || ( iRet != sizeof( Int32 ) ) )
        {
        retval = -1;
        }


    return ( retval );
    }


/**
 * Reads the given number of bytes from the given socket.
 */
static Int32 NET_SocketRead( Int32 iSocket, void *ptr, Int32 iNumBytes ) 
    {
    Int32 iBytesLeft;
    Int32 iRet;
#ifndef NDEBUG
    char *tptr = ( char * ) ptr;
    int i;
#endif /* NDEBUG */

    /* We may have to do multiple read( ) calls here */
    iBytesLeft = iNumBytes;

    while ( iBytesLeft > 0 ) 
        {
        iRet = read( iSocket, ptr, iBytesLeft );

        /* If an error occured, return */
        if ( iRet < 0 ) 
            {
            return iRet;
            }
        else if ( iRet == 0 ) 
            {
            break; /* EOF */
            }

        iBytesLeft  -= iRet;
        ptr = ( Byte * ) ptr + iRet;
        }

#ifndef NDEBUG
    printf( "NET_SocketRead: %.4X", iNumBytes - iBytesLeft );
    for ( i = 0; i < ( iNumBytes - iBytesLeft ); i++ )
        {
        printf( " %2.2X", tptr[ i ] );
        }
    printf( "\n" );
#endif /* NDEBUG */

    /* Return the number of bytes read */
    return ( iNumBytes - iBytesLeft );
    }


/**
 * Write the given number of bytes to the given socket
 */
static Int32 NET_SocketWrite( Int32 iSocket, void *ptr, Int32 iNumBytes ) 
    {
    Int32 iBytesLeft;
    Int32 iRet;
#ifndef NDEBUG
    char *tptr = ( char * ) ptr;
    int i;
#endif /* NDEBUG */

    /* We may have to do multiple write( ) calls here */
    iBytesLeft = iNumBytes;

    while ( iBytesLeft > 0 ) 
        {
        iRet = write( iSocket, ptr, iBytesLeft );

        /* If an error occured, return */
        if ( iRet < 0 ) 
            {
            return ( -1 );
            }

        iBytesLeft  -= iRet;
        ptr = ( Byte * ) ptr + iRet;
        }

#ifndef NDEBUG
    printf( "NET_SocketWrite: %.4X", iNumBytes - iBytesLeft );
    for ( i = 0; i < ( iNumBytes - iBytesLeft ); i++ )
        {
        printf( " %2.2X", tptr[ i ] );
        }
    printf( "\n" );
#endif /* NDEBUG */

    return ( iNumBytes - iBytesLeft );
    }


/**
 * Free the memory allocated for the given message.
 *
 * \bug void pointer
 */
void NET_DeleteMessage( Msgs *pMsg ) 
    {
    /*
     * Sanity check
     */
    if ( pMsg == NULL )
        {
        return;
        }

    switch( pMsg->msgId ) 
        {
        case MSG_OLDREGISTERCLIENT:
            MEM_Free( pMsg->msgOldRegisterClient.strClientAddress );
            break;

        case MSG_REGISTERCLIENT:
            MEM_Free( pMsg->msgRegisterClient.strClientAddress );
            break;

        case MSG_SENDMESSAGE:
            MEM_Free( pMsg->msgSendMessage.strMessage );
            MEM_Free( pMsg->msgSendMessage.strDestination );
            break;

        case MSG_MESSAGEPACKET:
            MEM_Free( pMsg->msgMessagePacket.strMessage );
            break;

        case MSG_OBJSTRUPDATE:
            if ( pMsg->msgObjStrUpdate.strNewValue )
                {
                MEM_Free( pMsg->msgObjStrUpdate.strNewValue );
                }
            break;

        case MSG_NETMESSAGE:
            MEM_Free( pMsg->msgNetMessage.strMessage );
            break;

        case MSG_NETPOPUP:
            MEM_Free( pMsg->msgNetPopup.strMessage );
            break;

        case MSG_VERSION:
            MEM_Free( pMsg->msgVersion.strVersion );
            break;

        case MSG_NOMESSAGE:
        case MSG_OBJINTUPDATE:
        case MSG_FREEPLAYER:
        case MSG_TURNNOTIFY:
        case MSG_CLIENTIDENT:
        case MSG_REQUESTCARD:
        case MSG_ENTERSTATE:
        case MSG_ENDOFMISSION:
        case MSG_VICTORY:
        case MSG_DICEROLL:
        case MSG_PLACENOTIFY:
        case MSG_ATTACKNOTIFY:
        case MSG_MOVENOTIFY:
        case MSG_SPECIESIDENT:
        case MSG_FORCEEXCHANGECARDS:
        case MSG_EXCHANGECARDS:
        case MSG_CARDPACKET:
        case MSG_REPLYPACKET:
        case MSG_EXIT:
        case MSG_STARTGAME:
        case MSG_ENDTURN:
        case MSG_DEREGISTERCLIENT:
        case MSG_ALLOCPLAYER:
        case MSG_POPUPREGISTERBOX:
        case MSG_HELLO:
        case MSG_MISSION:
        case MSG_ENDOFGAME:
            break;

        default:
            assert( FALSE );
        }

    MEM_Free( pMsg );
    }


/**
 * Set socket options for better performance.
 */
void NET_SetCommLinkOptions( Int32 iSocket ) 
    {
#ifdef TCP_NODELAY
    {
    Int32 iOption = 1;

    if ( setsockopt( iSocket, IPPROTO_TCP, TCP_NODELAY,
                   ( char * ) &iOption, sizeof( iOption ) ) != 0 ) 
        {
        printf( "NETWORK: Warning -- couldn't set TCP_NODELAY on CommLink!\n" );
        }
    }
#endif
#ifdef SO_REUSEADDR
    {
    Int32 iOption = 1;

    if ( setsockopt( iSocket, SOL_SOCKET, SO_REUSEADDR,
                   ( char * ) &iOption, sizeof( iOption ) ) != 0 ) 
        {
        printf( "NETWORK: Warning -- couldn't set SO_REUSEADDR on CommLink!\n" );
        }
    }
#endif
    }


/**
 * Convert the given message to a human readable string for debugging.
 */
CString NET_MessageToString( const Msgs *pMsg ) 
    {
    /*
     * Sanity check
     */
    if ( pMsg == NULL )
        {
        return NULL;
        }

    switch ( pMsg->msgId ) 
        {
        case MSG_NOMESSAGE:
            snprintf( strLogging, sizeof( strLogging ) , "MsgNoMessage( ) " );
            break;

        case MSG_OLDREGISTERCLIENT:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgOldRegisterClient( strClientAddress=\"%s\" ) ",
                      pMsg->msgOldRegisterClient.strClientAddress );
            break;

        case MSG_REGISTERCLIENT:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgRegisterClient( strClientAddress=\"%s\", "
                      " iClientType=%s ) ",
                      pMsg->msgRegisterClient.strClientAddress,
                      pMsg->msgRegisterClient.iClientType ==
                      CLIENT_NORMAL ? "CLIENT_NORMAL" :
                      pMsg->msgRegisterClient.iClientType ==
                      CLIENT_AI ? "CLIENT_AI" :
                      pMsg->msgRegisterClient.iClientType ==
                      CLIENT_STRICTOBSERVER ? "CLIENT_STRICTOBSERVER" :
                      "*Unknown*" );
            break;

        case MSG_EXCHANGECARDS:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgExchangeCards( piCards={%d, %d, %d} ) ",
                      pMsg->msgExchangeCards.piCards[ 0 ],
                      pMsg->msgExchangeCards.piCards[ 1 ],
                      pMsg->msgExchangeCards.piCards[ 2 ] );
            break;

        case MSG_CARDPACKET:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgCardPacket( iPlayer=%d, iCard=%d ) ",
                      pMsg->msgCardPacket.iPlayer,
                      pMsg->msgCardPacket.cdCard );
            break;

        case MSG_REPLYPACKET:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgReplyPacket( iReply=%d ) ",
                      pMsg->msgReplyPacket.iReply );
            break;

        case MSG_SENDMESSAGE:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgSendMessage( strMessage=\"%s\", "
                      "strDestination=\"%s\" ) ",
                      pMsg->msgSendMessage.strMessage,
                      pMsg->msgSendMessage.strDestination );
            break;

        case MSG_MESSAGEPACKET:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgMessagePacket( strMessage=\"%s\", "
                      "iFrom=\"%d\", iTo=%d ) ",
                      pMsg->msgMessagePacket.strMessage,
                      pMsg->msgMessagePacket.iFrom,
                      pMsg->msgMessagePacket.iTo );
            break;

        case MSG_TURNNOTIFY:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgTurnNotify( iPlayer=%d, iClient=%d ) ",
                      pMsg->msgTurnNotify.iPlayer,
                      pMsg->msgTurnNotify.iClient );
            break;

        case MSG_CLIENTIDENT:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgClientIdent( iClientID=%d ) ",
                      pMsg->msgClientIdent.iClientID );
            break;

        case MSG_REQUESTCARD:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgRequestCard( iPlayer=%d ) ",
                      pMsg->msgRequestCard.iPlayer );
            break;

        case MSG_ENTERSTATE:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgEnterState( iState=%d ) ",
                      pMsg->msgEnterState.iState );
            break;

        case MSG_OBJSTRUPDATE:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "ObjStrUpdate( iField=%d, iIndex=[%d, %d], "
                      "strNewValue=\"%s\" ) ",
                      pMsg->msgObjStrUpdate.iField,
                      pMsg->msgObjStrUpdate.iIndex1,
                      pMsg->msgObjStrUpdate.iIndex2,
                      pMsg->msgObjStrUpdate.strNewValue );
            break;

        case MSG_OBJINTUPDATE:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "ObjIntUpdate( iField=%d, iIndex=[%d, %d], "
                      "iNewValue=%d ) ",
                      pMsg->msgObjIntUpdate.iField,
                      pMsg->msgObjIntUpdate.iIndex1,
                      pMsg->msgObjIntUpdate.iIndex2,
                      pMsg->msgObjIntUpdate.iNewValue );
            break;

        case MSG_FREEPLAYER:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgFreePlayer( iPlayer=%d ) ",
                      pMsg->msgFreePlayer.iPlayer );
            break;

        case MSG_NETMESSAGE:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgNetMessage( strMessage=\"%s\" ) ",
                      pMsg->msgNetMessage.strMessage );
            break;

        case MSG_NETPOPUP:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgNetPopup( strMessage=\"%s\" ) ",
                      pMsg->msgNetPopup.strMessage );
            break;

        case MSG_ENDOFMISSION:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgEndOfMission( iWinner=%d, iTyp=%d, iNum1=%d, iNum2=%d ) ",
                      pMsg->msgEndOfMission.iWinner,
                      pMsg->msgEndOfMission.iTyp,
                      pMsg->msgEndOfMission.iNum1,
                      pMsg->msgEndOfMission.iNum2 );
            break;

        case MSG_VICTORY:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgVictory( iWinner=%d ) ",
                      pMsg->msgVictory.iWinner );
            break;

        case MSG_ENDOFGAME:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgEndOfGame( ) " );
            break;

        case MSG_DICEROLL:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgDiceRoll( iDefendingPlayer=%d, "
                      "pAttackDice={%d, %d, %d}, pDefendDice={%d, %d, %d} ) ",
                      pMsg->msgDiceRoll.iDefendingPlayer,
                      pMsg->msgDiceRoll.pAttackDice[ 0 ],
                      pMsg->msgDiceRoll.pAttackDice[ 1 ],
                      pMsg->msgDiceRoll.pAttackDice[ 2 ],
                      pMsg->msgDiceRoll.pDefendDice[ 0 ],
                      pMsg->msgDiceRoll.pDefendDice[ 1 ],
                      pMsg->msgDiceRoll.pDefendDice[ 2 ] );
            break;

        case MSG_PLACENOTIFY:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgPlaceNotify( iCountry=%d ) ",
                      pMsg->msgPlaceNotify.iCountry );
            break;

        case MSG_ATTACKNOTIFY:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgAttackNotify( iSrcCountry=%d, iDstCountry=%d ) ",
                      pMsg->msgAttackNotify.iSrcCountry,
                      pMsg->msgAttackNotify.iDstCountry );
            break;

        case MSG_MOVENOTIFY:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgMoveNotify( iSrcCountry=%d, iDstCountry=%d ) ",
                      pMsg->msgMoveNotify.iSrcCountry,
                      pMsg->msgMoveNotify.iDstCountry );
            break;

        case MSG_VERSION:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgVersion( strVersion=\"%s\" ) ",
                      pMsg->msgVersion.strVersion );
            break;

        case MSG_SPECIESIDENT:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgSpeciesIdent( iSpeciesID=%d ) ",
                      pMsg->msgSpeciesIdent.iSpeciesID );
            break;

        case MSG_FORCEEXCHANGECARDS:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgForceExchangeCards( iPlayer=%d ) ",
                      pMsg->msgForceExchangeCards.iPlayer );
            break;

        case MSG_EXIT:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgExit( ) " );
            break;

        case MSG_STARTGAME:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgStartGame( ) " );
            break;

        case MSG_ENDTURN:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgEndTurn( ) " );
            break;

        case MSG_DEREGISTERCLIENT:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgDeregisterClient( ) " );
            break;

        case MSG_ALLOCPLAYER:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgAllocPlayer( ) " );
            break;

        case MSG_POPUPREGISTERBOX:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgPopupRegisterBox( ) " );
            break;

        case MSG_HELLO:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgHello( ) " );
            break;

        case MSG_MISSION:
            snprintf( strLogging,
                      sizeof( strLogging ),
                      "MsgMission( ) " );
            break;

        default:
            assert( FALSE );
        }

    return strLogging;
    }

/* End of file */
