/*
 *   Xfrisk - The classic board game for X
 *   Copyright (C) 1993-1999 Elan Feingold (elan@aetherworks.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   $Id: network.h,v 1.18 2003/05/13 23:32:56 sketch Exp $
 */

#ifndef NETWORK_H
#define NETWORK_H

#include <sys/types.h>

#include "types.h"

/**
 * Message ID numbers
 * When adding new messages to the protocol, please reuse ID numbers
 * marked as not used in this enumeration. Don't forget to check that
 * you either don't break backward compatibility or that you make it
 * impossible for older clients to communicate with a newer server.
 */
typedef enum
    {
    MSG_UNUSED_1            = 0x00,
    MSG_OLDREGISTERCLIENT   = 0x01,
    MSG_UNUSED_2            = 0x02,
    MSG_EXCHANGECARDS       = 0x03,
    MSG_REQUESTCARD         = 0x04,
    MSG_CARDPACKET          = 0x05,
    MSG_REPLYPACKET         = 0x06,
    MSG_SENDMESSAGE         = 0x07,
    MSG_MESSAGEPACKET       = 0x08,
    MSG_EXIT                = 0x09,
    MSG_STARTGAME           = 0x0A,
    MSG_TURNNOTIFY          = 0x0B,
    MSG_ENDTURN             = 0x0C,
    MSG_CLIENTIDENT         = 0x0D,
    MSG_ENTERSTATE          = 0x0E,
    MSG_UNUSED_3            = 0x0F,
    MSG_ENDOFGAME           = 0x10,
    MSG_OBJSTRUPDATE        = 0x11,
    MSG_OBJINTUPDATE        = 0x12,
    MSG_UNUSED_4            = 0x13,
    MSG_UNUSED_5            = 0x14,
    MSG_UNUSED_6            = 0x15,
    MSG_DEREGISTERCLIENT    = 0x16,
    MSG_UNUSED_7            = 0x17,
    MSG_UNUSED_8            = 0x18,
    MSG_ALLOCPLAYER         = 0x19,
    MSG_UNUSED_9            = 0x1A,
    MSG_UNUSED_10           = 0x1B,
    MSG_UNUSED_11           = 0x1C,
    MSG_UNUSED_12           = 0x1D,
    MSG_UNUSED_13           = 0x1E,
    MSG_UNUSED_14           = 0x1F,
    MSG_FREEPLAYER          = 0x20,
    MSG_NETMESSAGE          = 0x21,
    MSG_NETPOPUP            = 0x22,
    MSG_POPUPREGISTERBOX    = 0x23,
    MSG_DICEROLL            = 0x24,
    MSG_PLACENOTIFY         = 0x25,
    MSG_ATTACKNOTIFY        = 0x26,
    MSG_MOVENOTIFY          = 0x27,
    MSG_POLLCLIENTS         = 0x28,
    MSG_UNUSED_15           = 0x29,
    MSG_HELLO               = 0x2A,
    MSG_VERSION             = 0x2B,
    MSG_SPECIESIDENT        = 0x2C,
    MSG_MISSION             = 0x2D,
    MSG_ENDOFMISSION        = 0x2E,
    MSG_VICTORY             = 0x2F,
    MSG_FORCEEXCHANGECARDS  = 0x30,
    MSG_NUM_CONTIGUOUS      = 0x31,
    MSG_REGISTERCLIENT      = 0x99,
    MSG_NOMESSAGE           = 0xDEAD,
    MSG_MAX_ID              = 0xFFFFF
    } MsgId;

/**
 * Types of clients to connect
 */
typedef enum
    {
    CLIENT_NORMAL           = 0,
    CLIENT_AI               = 1,
    CLIENT_STRICTOBSERVER   = 2 /* TBD, for regression testing */
    } ClientTypes;

/**
 * Special message source IDs
 */
typedef enum
    {
    FROM_SERVER = -1,
    FROM_UNKNOW = -2
    } MsgSpecialSourceIDs;

/**
 * Special message destination IDs
 */
typedef enum
    {
    DST_ALLPLAYERS  = -1,
    DST_ALLBUTME    = -2,
    DST_OTHER       = -3
    } MsgSpecialDestinationIDs;


/*
 * The individual message types
 *
 * NOTE: Each message must start with the message primitive!
 */

/**
 * Register a client (old message)
 */
typedef struct
    {
    MsgId   msgId;
    CString strClientAddress;
    } MsgOldRegisterClient;

/**
 * Register a client
 */
typedef struct
    {
    MsgId   msgId;
    CString strClientAddress;
    Int32   iClientType;  /* Flags */
    } MsgRegisterClient;

/**
 * Cards to be exchanged
 */
typedef struct
    {
    MsgId   msgId;
    Int32   piCards[3];
    } MsgExchangeCards;

/**
 * Request a card
 * \bug Meaning is obscure
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iPlayer;
    } MsgRequestCard;

/**
 * A card packet
 * \bug Meaning is obscure
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iPlayer;
    Int32   cdCard;
    } MsgCardPacket;

/**
 * Reply packet
 * \bug Meaning is obscure
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iReply;
    } MsgReplyPacket;

/**
 * Send message
 * \bug Meaning is obscure
 */
typedef struct
    {
    MsgId   msgId;
    CString strMessage;
    CString strDestination;
    } MsgSendMessage;

/**
 * Message packet
 */
typedef struct
    {
    MsgId   msgId;
    CString strMessage;
    Int32   iFrom;
    Int32   iTo;
    } MsgMessagePacket;

/**
 * Turn notification
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iPlayer;
    Int32   iClient;
    } MsgTurnNotify;

/**
 * Client identification
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iClientID;
    } MsgClientIdent;

/**
 * State change notification
 * \bug This is not really a good thing...
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iState;
    } MsgEnterState;

/**
 * Object string property update
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iField;
    Int32   iIndex1;
    Int32   iIndex2;
    CString strNewValue;
    } MsgObjStrUpdate;

/**
 * Object integer property update
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iField;
    Int32   iIndex1;
    Int32   iIndex2;
    Int32   iNewValue;
    } MsgObjIntUpdate;

/**
 * Free player entry notification
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iPlayer;
    } MsgFreePlayer;

/**
 * Message from server to clients
 */
typedef struct
    {
    MsgId   msgId;
    CString strMessage;
    } MsgNetMessage;

/**
 * Popup message from server to clients
 */
typedef struct
    {
    MsgId   msgId;
    CString strMessage;
    } MsgNetPopup;

/**
 * Dice roll notification
 *
 * \bug Should be from server to clients to prevent cheating
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iDefendingPlayer;
    Int32   pAttackDice[ 3 ];
    Int32   pDefendDice[ 3 ];
    } MsgDiceRoll;

/**
 * Army placement notification
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iCountry;
    } MsgPlaceNotify;

/**
 * Attack notification
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iSrcCountry;
    Int32   iDstCountry;
    } MsgAttackNotify;

/**
 * Army movement notification
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iSrcCountry;
    Int32   iDstCountry;
    } MsgMoveNotify;

/**
 * Poll?
 * \bug Meaning is obscure
 */
typedef struct
    {
    MsgId   msgId;
    CString strPollQuestion;
    } MsgPollClients;

/**
 * Version information
 */
typedef struct
    {
    MsgId   msgId;
    CString strVersion;
    } MsgVersion;

/**
 * Species identification
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iSpeciesID;
    } MsgSpeciesIdent;

/**
 * Mission end notification
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iWinner;
    Int32   iTyp;
    Int32   iNum1;
    Int32   iNum2;
    } MsgEndOfMission;

/**
 * Victory indication
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iWinner;
    } MsgVictory;

/**
 * Forced card exchange
 */
typedef struct
    {
    MsgId   msgId;
    Int32   iPlayer;
    } MsgForceExchangeCards;


/**
 * Union of all message types
 */
typedef union
    {
    MsgId                   msgId;
    MsgOldRegisterClient    msgOldRegisterClient;
    MsgRegisterClient       msgRegisterClient;
    MsgExchangeCards        msgExchangeCards;
    MsgRequestCard          msgRequestCard;
    MsgCardPacket           msgCardPacket;
    MsgReplyPacket          msgReplyPacket;
    MsgSendMessage          msgSendMessage;
    MsgMessagePacket        msgMessagePacket;
    MsgTurnNotify           msgTurnNotify;
    MsgClientIdent          msgClientIdent;
    MsgEnterState           msgEnterState;
    MsgObjStrUpdate         msgObjStrUpdate;
    MsgObjIntUpdate         msgObjIntUpdate;
    MsgFreePlayer           msgFreePlayer;
    MsgNetMessage           msgNetMessage;
    MsgNetPopup             msgNetPopup;
    MsgDiceRoll             msgDiceRoll;
    MsgPlaceNotify          msgPlaceNotify;
    MsgAttackNotify         msgAttackNotify;
    MsgMoveNotify           msgMoveNotify;
    MsgPollClients          msgPollClients;
    MsgVersion              msgVersion;
    MsgSpeciesIdent         msgSpeciesIdent;
    MsgEndOfMission         msgEndOfMission;
    MsgVictory              msgVictory;
    MsgForceExchangeCards   msgForceExchangeCards;
    } Msgs;

Int32    NET_SendMessage( Int32 iSocket, const Msgs *pMsg );
Int32    NET_RecvMessage( Int32 iSocket, Msgs **ppMsg );
void     NET_DeleteMessage( Msgs *pMsg );
void     NET_SetCommLinkOptions( Int32 iSocket );
CString  NET_MessageToString( const Msgs *pMsg );

#endif /* NETWORK_H */

/* End of file */
