/*
 *   Xfrisk - The classic board game for X
 *   Copyright (C) 1993-1999 Elan Feingold (elan@aetherworks.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   $Id: riskgame.h,v 1.18 2003/07/14 01:17:10 morphy Exp $
 */

/** \file
 * Interface definitions for the RISK game distributed object.
 *
 * \bug Contains a LOT of hardcoded information
 */

#ifndef RISKGAME_H
#define RISKGAME_H

#include <stdio.h>

#include "network.h"
#include "types.h"

#ifdef __GNUC__
#define UNUSED( x )                             \
    ( ( void ) x )
#else
#define UNUSED( x )
#endif

/**
 * Game states
 */
typedef enum
    {
    GAME_REGISTERING = 0,
    GAME_FORTIFYING  = 1,
    GAME_PLAYING     = 2
    } GAME_States;

extern GAME_States  iGameState;

/* Fields of the RiskGame object.  There MUST NOT be any gaps in the
 * numeric sequence because of how the replication works.
 */

typedef enum
    {
    PLR_ATTACKMODE      = 0,   /* Remembers the attack mode of the player */
    PLR_MSGDSTMODE      = 1,   /* Remembers the last message destination */
    PLR_DICEMODE        = 2,   /* Remembers the last die mode */
    PLR_STATE           = 3,   /* Either TRUE = Alive, FALSE = Dead */
    PLR_CLIENT          = 4,   /* The client of the player */
    PLR_NUMCOUNTRIES    = 5,   /* Number of territories the player owns */
    PLR_NUMARMIES       = 6,   /* Number of armies the player holds */
    PLR_NUMCARDS        = 7,   /* Number of cards the player holds */
    PLR_NAME            = 8,   /* Name of the player */
    PLR_COLORSTRING     = 9,   /* Name of the color of the player */
    PLR_CARD            = 10,  /* A card of the player */
    PLR_SPECIES         = 11,  /* What is the player */
    PLR_ALLOCATION      = 12,  /* State of allocation of player */
    PLR_MISSION         = 13,  /* Mission's type of player */
    PLR_MISSION1        = 14,  /* Mission's number1 of player */
    PLR_MISSION2        = 15,  /* Mission's number2 of player */
    CNT_NAME            = 16,
    CNT_CONTINENT       = 17,
    CNT_NUMARMIES       = 18,
    CNT_OWNER           = 19,
    CNT_ADJCOUNTRY      = 20,
    CNT_TEXTX           = 21,
    CNT_TEXTY           = 22,
    CON_NAME            = 23,
    CON_VALUE           = 24,
    CON_NUMCOUNTRIES    = 25,
    SPE_NAME            = 26,
    SPE_VERSION         = 27,
    SPE_DESCRIPTION     = 28,
    SPE_AUTHOR          = 29,
    SPE_CLIENT          = 30,
    SPE_ALLOCATION      = 31
    } RiskGameFields;

/* Port at which the clients can find the server */
#define RISK_PORT       5324

/* Amount of time (in ms) for graphical notification */
#define NOTIFY_TIME     750

/* Limits */
#define MAX_CLIENTS     32
#define MAX_CARDS       10
#define MAX_PLAYERS     12
#define MAX_COLORS      256

/* This may be a static value (i.e. OPEN_MAX), or a dynamic value,
 * as in OSF/1, accessed through sysconf().  I'm not sure that all 
 * systems have these, so we'll just guess.  I don't think this is
 * exceptionally evil, since if we run out of descriptors, the socket
 * or accept calls will fail.
 */
#define MAX_DESCRIPTORS     128

/* Numbers */
#define NUM_CONTINENTS      6
#define NUM_COUNTRIES       42
#define NUM_OTHERCOLORS     4 /* Dice*2, Player Turn, ColorEdit */
#define NUM_CARDS           (NUM_COUNTRIES+2)
#define NUM_FORTIFY_ARMIES  1

/* Continents */
typedef enum
    {
    CNT_NORTHAMERICA        = 0,
    CNT_SOUTHAMERICA        = 1,
    CNT_AFRICA              = 2,
    CNT_AUSTRALIA           = 3,
    CNT_ASIA                = 4,
    CNT_EUROPE              = 5
    } RISK_Continents;

/* Missions */
typedef enum
    {
    NO_MISSION              = 0,
    CONQUIER_WORLD          = 1,
    CONQUIER_Nb_COUNTRY     = 2,
    CONQUIER_TWO_CONTINENTS = 3,
    KILL_A_PLAYER           = 4
    } RISK_Missions;

/* Macros */
#ifndef MAX
#define MAX( a, b )                             \
    ( ( a ) > ( b ) ? ( a ) : ( b ) )
#endif

#ifndef MIN
#define MIN( a, b )                             \
    ( ( a ) > ( b ) ? ( b ) : ( a ) )
#endif

#ifndef TRUE
#define TRUE   1
#endif

#ifndef FALSE
#define FALSE  0
#endif

/* States in risk state machine */
typedef enum
    {
    STATE_REGISTER  = 0,
    STATE_FORTIFY   = 1,
    STATE_PLACE     = 2,
    STATE_ATTACK    = 3,
    STATE_MOVE      = 4
    } RISK_GameStates;

/* Defines for widgets */
typedef enum
    {
    ATTACK_ONE      = 0,
    ATTACK_TWO      = 1,
    ATTACK_THREE    = 2,
    ATTACK_AUTO     = 3
    } AttackChoices;

typedef enum
    {
    ACTION_PLACE    = 0,
    ACTION_ATTACK   = 1,
    ACTION_DOORDIE  = 2,
    ACTION_MOVE     = 3
    } RISK_Actions;

/* The only known species at the beginning */
#define SPECIES_HUMAN       0

/* A PERSISTANT object will maintain a copy of itself on secondary storage.  
 * Every time the object updates a field, it also updates the field on disk.
 */

#define MODE_PERSISTANT     8

/* States in allocation for players and species */
#define ALLOC_NONE          0
#define ALLOC_INPROGRESS    1
#define ALLOC_COMPLETE      2

/* States for players */
#define PLAYER_DEAD         0
#define PLAYER_ALIVE        1

/* These functions support all of the functionality of the RiskGame structure,
 * in order for it to work as a fully distributed object.  The functions
 * are split into two catagories, those that receive input from a remote 
 * source (through a message), those that receive input from the local
 * client, which is sent back to the server to broadcast to all the other
 * clients, and those that return fields of the object to the local client.
 * This file serves the functionality of a C++ object -- I would use GCC but
 * its C++ side is not stable enough.  Hold your breath, though :)
 */

/* Constructor */

void    RISK_InitObject( void ( *ReplicateCallback )( const Msgs *, Int32, Int32 ),
                         void ( *PreViewCallback )( const Msgs * ),
                         void ( *PostViewCallback )( const Msgs * ),
                         void ( *FailureCallback )( CString, Int32 ),
                         FILE *hDebug );

#define MESS_INCOMING       0
#define MESS_OUTGOING       1

/* For persistance */

void     RISK_SaveObject( CString strFileName );
void     RISK_LoadObject( CString strFileName );


/* Utilities */

Int32    RISK_GetNthLivePlayer( Int32 iIndex );
Int32    RISK_GetNthPlayer( Int32 iIndex );
Int32    RISK_GetNthPlayerAtClient( Int32 iClient, Int32 iIndex );
Int32    RISK_GetNumPlayersOfClient( Int32 iClient );
Int32    RISK_GetNumLivePlayersOfClient( Int32 iClient );
void     RISK_ObjectFailure( CString strReason, Int32 iCommLink );

/* For the failure routine */
#define SERVER -1


/* Reliable communication */

Int32    RISK_SendMessage( Int32 iDest, const Msgs *pMessage );
Int32    RISK_ReceiveMessage( Int32 iSource, Msgs **ppMessage );

Int32    RISK_SendSyncMessage( Int32 iCommLink,
                               const Msgs *pMessage,
                               MsgId iReturnMessType,
                               void ( *CBK_MessageReceived )( const Msgs * ) );

/* Methods for the distributed object. */

void     RISK_SelectiveReplicate( Int32 iClientToServerSocket,
                                  Int32 iField, 
                                  Int32 iIndex1,
                                  Int32 iIndex2 );

void     RISK_ResetObj( void );
void     RISK_ResetGame( void );
void     RISK_ProcessMessage( const Msgs *pMessage );

void     RISK_SetSpeciesOfPlayer( Int32 iPlayer, Int32 iSpecies );
void     RISK_SetAttackModeOfPlayer( Int32 iPlayer, Int32 iMode );
void     RISK_SetDiceModeOfPlayer( Int32 iPlayer, Int32 iMode );
void     RISK_SetMsgDstModeOfPlayer( Int32 iPlayer, Int32 iMode );
void     RISK_SetStateOfPlayer( Int32 iPlayer, Int32 iState );
void     RISK_SetClientOfPlayer( Int32 iPlayer, Int32 iClient );
void     RISK_SetAllocationStateOfPlayer( Int32 iPlayer, Int32 iState );
void     RISK_SetNumCountriesOfPlayer( Int32 iCountry, Int32 iNumCountries );
void     RISK_SetNumArmiesOfPlayer( Int32 iCountry, Int32 iNumArmies );
void     RISK_SetNumCardsOfPlayer( Int32 iPlayer, Int32 iNumCards );
void     RISK_SetNameOfPlayer( Int32 iPlayer, CString strName );
void     RISK_SetColorCStringOfPlayer( Int32 iPlayer, CString strColor );

void     RISK_SetCardOfPlayer( Int32 iPlayer,
                               Int32 iCard,
                               Int32 iValue );

void     RISK_SetMissionTypeOfPlayer( Int32 iPlayer, Int16 typ );
void     RISK_SetMissionNumberOfPlayer( Int32 iPlayer, Int32 n );
void     RISK_SetMissionContinent1OfPlayer( Int32 iPlayer, Int32 n );
void     RISK_SetMissionContinent2OfPlayer( Int32 iPlayer, Int32 n );
void     RISK_SetMissionMissionPlayerToKillOfPlayer( Int32 iPlayer, Int32 n );
void     RISK_SetMissionPlayerIsKilledOfPlayer( Int32 iPlayer, Flag boool );
void     RISK_SetNameOfCountry( Int32 iCountry, CString strName );
void     RISK_SetContinentOfCountry( Int32 iCountry, Int32 iContinent );
void     RISK_SetOwnerOfCountry( Int32 iCountry, Int32 iOwner );
void     RISK_SetNumArmiesOfCountry( Int32 iCountry, Int32 iNumArmies );

void     RISK_SetAdjCountryOfCountry( Int32 iCountry,
                                      Int32 iIndex, 
                                      Int32 OtherC );

Int32    RISK_GetSpeciesOfPlayer( Int32 iPlayer );
Int32    RISK_GetDiceModeOfPlayer( Int32 iPlayer );
Int32    RISK_GetMsgDstModeOfPlayer( Int32 iPlayer );
Int32    RISK_GetAttackModeOfPlayer( Int32 iPlayer );
Int32    RISK_GetStateOfPlayer( Int32 iPlayer );
Int32    RISK_GetClientOfPlayer( Int32 iPlayer );
Int32    RISK_GetAllocationStateOfPlayer( Int32 iPlayer );
Int32    RISK_GetNumCountriesOfPlayer( Int32 iPlayer );
Int32    RISK_GetNumArmiesOfPlayer( Int32 iPlayer );
Int32    RISK_GetNumCardsOfPlayer( Int32 iPlayer );
CString  RISK_GetNameOfPlayer( Int32 iPlayer );
CString  RISK_GetColorCStringOfPlayer( Int32 iPlayer );
Int32    RISK_GetCardOfPlayer( Int32 iPlayer, Int32 iCard );
Int16    RISK_GetMissionTypeOfPlayer( Int32 iPlayer );
Int32    RISK_GetMissionNumberOfPlayer( Int32 iPlayer );
Int32    RISK_GetMissionContinent1OfPlayer( Int32 iPlayer );
Int32    RISK_GetMissionContinent2OfPlayer( Int32 iPlayer );
Int32    RISK_GetMissionPlayerToKillOfPlayer( Int32 iPlayer );
Flag     RISK_GetMissionIsPlayerKilledOfPlayer( Int32 iPlayer );
CString  RISK_GetNameOfCountry( Int32 iCountry );
Int32    RISK_GetContinentOfCountry( Int32 iCountry );
Int32    RISK_GetNumArmiesOfCountry( Int32 iCountry );
Int32    RISK_GetOwnerOfCountry( Int32 iCountry );
Int32    RISK_GetAdjCountryOfCountry( Int32 iCountry, Int32 iIndex );
Int32    RISK_GetTextXOfCountry( Int32 iCountry );
Int32    RISK_GetTextYOfCountry( Int32 iCountry );
Int32    RISK_GetValueOfContinent( Int32 iContinent );
CString  RISK_GetNameOfContinent( Int32 iContinent );
Int32    RISK_GetNumCountriesOfContinent( Int32 iContinent );
Int32    RISK_GetNumLivePlayers( void );
Int32    RISK_GetNumPlayers( void );

/* The first (should-be-a) composite object in Xfrisk */
CString  RISK_GetNameOfSpecies( Int32 iHandle );
Int32    RISK_GetClientOfSpecies( Int32 iHandle );
CString  RISK_GetAuthorOfSpecies( Int32 iHandle );
CString  RISK_GetDescriptionOfSpecies( Int32 iHandle );
CString  RISK_GetVersionOfSpecies( Int32 iHandle );
Int32    RISK_GetAllocationStateOfSpecies( Int32 iHandle );
Int32    RISK_GetNumSpecies( void );

void     RISK_SetNameOfSpecies( Int32 iHandle, CString strName );
void     RISK_SetClientOfSpecies( Int32 iHandle, Int32 iClient );
void     RISK_SetAuthorOfSpecies( Int32 iHandle, CString strAuthor );
void     RISK_SetVersionOfSpecies( Int32 iHandle, CString strVersion );
void     RISK_SetDescriptionOfSpecies( Int32 iHandle, CString strDescription );
void     RISK_SetAllocationStateOfSpecies( Int32 iHandle, Int32 iState );
void     RISK_SetNumSpecies( Int32 iNumSpecies );

void     RISK_ReplicateRegistrationData( Int32 iCommLinkDest );
void     RISK_ReplicateAllData( Int32 iCommLinkDest );
void     RISK_HandleALLOCPLAYER( Int32 iClient, Int32 iPlayer );

#endif /* RISKGAME_H */

/* End of file */
